import json
from uuid import uuid4

import pytest

from labelbox.data.serialization.ndjson.converter import NDJsonConverter


def test_relationship():
    with open("tests/data/assets/ndjson/relationship_import.json", "r") as file:
        data = json.load(file)

    res = list(NDJsonConverter.deserialize(data))
    res = list(NDJsonConverter.serialize(res))
    assert len(res) == len(data)

    res_relationship_annotation, res_relationship_second_annotation = [
        annot for annot in res if "relationship" in annot
    ]
    res_source_and_target = [
        annot for annot in res if "relationship" not in annot
    ]
    assert res_relationship_annotation

    assert res_relationship_annotation["relationship"]["source"] in [
        annot["uuid"] for annot in res_source_and_target
    ]
    assert res_relationship_annotation["relationship"]["target"] in [
        annot["uuid"] for annot in res_source_and_target
    ]

    assert res_relationship_second_annotation
    assert res_relationship_second_annotation["relationship"][
        "source"] != res_relationship_annotation["relationship"]["source"]
    assert res_relationship_second_annotation["relationship"][
        "target"] != res_relationship_annotation["relationship"]["target"]
    assert res_relationship_second_annotation["relationship"]["source"] in [
        annot["uuid"] for annot in res_source_and_target
    ]
    assert res_relationship_second_annotation["relationship"]["target"] in [
        annot["uuid"] for annot in res_source_and_target
    ]


def test_relationship_nonexistent_object():
    with open("tests/data/assets/ndjson/relationship_import.json", "r") as file:
        data = json.load(file)

    relationship_annotation = data[2]
    source_uuid = relationship_annotation["relationship"]["source"]
    target_uuid = str(uuid4())
    relationship_annotation["relationship"]["target"] = target_uuid
    error_msg = f"Relationship object refers to nonexistent object with UUID '{source_uuid}' and/or '{target_uuid}'"

    with pytest.raises(ValueError, match=error_msg):
        list(NDJsonConverter.deserialize(data))


def test_relationship_duplicate_uuids():
    with open("tests/data/assets/ndjson/relationship_import.json", "r") as file:
        data = json.load(file)

    source, target = data[0], data[1]
    target["uuid"] = source["uuid"]
    error_msg = f"UUID '{source['uuid']}' is not unique"

    with pytest.raises(AssertionError, match=error_msg):
        list(NDJsonConverter.deserialize(data))



from types import SimpleNamespace

import pytest

from labelbox.data.annotation_types import ClassificationAnnotation, ObjectAnnotation
from labelbox.data.annotation_types import Polygon, Point


class NameSpace(SimpleNamespace):

    def __init__(self, predictions, ground_truths, expected):
        super(NameSpace, self).__init__(predictions=predictions,
                                        ground_truths=ground_truths,
                                        expected=expected)


@pytest.fixture
def different_classes():
    return [
        NameSpace(predictions=[
            ObjectAnnotation(name="cat",
                             value=Polygon(points=[
                                 Point(x=0, y=0),
                                 Point(x=10, y=0),
                                 Point(x=10, y=10),
                                 Point(x=0, y=10)
                             ]))
        ],
                  ground_truths=[
                      ObjectAnnotation(name="dog",
                                       value=Polygon(points=[
                                           Point(x=0, y=0),
                                           Point(x=10, y=0),
                                           Point(x=10, y=10),
                                           Point(x=0, y=10)
                                       ]))
                  ],
                  expected={
                      'cat': 0,
                      'dog': 0
                  })
    ]


@pytest.fixture
def one_overlap_class():
    return [
        NameSpace(predictions=[
            ObjectAnnotation(name="cat",
                             value=Polygon(points=[
                                 Point(x=0, y=0),
                                 Point(x=10, y=0),
                                 Point(x=10, y=10),
                                 Point(x=0, y=10)
                             ])),
            ObjectAnnotation(name="dog",
                             value=Polygon(points=[
                                 Point(x=0, y=0),
                                 Point(x=5, y=0),
                                 Point(x=5, y=5),
                                 Point(x=0, y=5)
                             ]))
        ],
                  ground_truths=[
                      ObjectAnnotation(name="dog",
                                       value=Polygon(points=[
                                           Point(x=0, y=0),
                                           Point(x=10, y=0),
                                           Point(x=10, y=10),
                                           Point(x=0, y=10)
                                       ]))
                  ],
                  expected={
                      'dog': 0.25,
                      'cat': 0.
                  }),
        NameSpace(predictions=[
            ObjectAnnotation(name="dog",
                             value=Polygon(points=[
                                 Point(x=0, y=0),
                                 Point(x=5, y=0),
                                 Point(x=5, y=5),
                                 Point(x=0, y=5)
                             ]))
        ],
                  ground_truths=[
                      ObjectAnnotation(name="dog",
                                       value=Polygon(points=[
                                           Point(x=0, y=0),
                                           Point(x=10, y=0),
                                           Point(x=10, y=10),
                                           Point(x=0, y=10)
                                       ])),
                      ObjectAnnotation(name="cat",
                                       value=Polygon(points=[
                                           Point(x=0, y=0),
                                           Point(x=10, y=0),
                                           Point(x=10, y=10),
                                           Point(x=0, y=10)
                                       ]))
                  ],
                  expected={
                      'dog': 0.25,
                      'cat': 0.
                  })
    ]


@pytest.fixture
def empty_annotations():
    return [
        NameSpace(predictions=[], ground_truths=[], expected={}),
        NameSpace(predictions=[],
                  ground_truths=[
                      ObjectAnnotation(name="dog",
                                       value=Polygon(points=[
                                           Point(x=0, y=0),
                                           Point(x=10, y=0),
                                           Point(x=10, y=10),
                                           Point(x=0, y=10)
                                       ])),
                      ObjectAnnotation(name="cat",
                                       value=Polygon(points=[
                                           Point(x=0, y=0),
                                           Point(x=10, y=0),
                                           Point(x=10, y=10),
                                           Point(x=0, y=10)
                                       ]))
                  ],
                  expected={
                      'dog': 0.,
                      'cat': 0.
                  }),
        NameSpace(predictions=[
            ObjectAnnotation(name="dog",
                             value=Polygon(points=[
                                 Point(x=0, y=0),
                                 Point(x=10, y=0),
                                 Point(x=10, y=10),
                                 Point(x=0, y=10)
                             ])),
            ObjectAnnotation(name="cat",
                             value=Polygon(points=[
                                 Point(x=0, y=0),
                                 Point(x=10, y=0),
                                 Point(x=10, y=10),
                                 Point(x=0, y=10)
                             ]))
        ],
                  ground_truths=[],
                  expected={
                      'dog': 0.,
                      'cat': 0.
                  })
    ]



import uuid
from labelbox import parser
import pytest
"""
- Here we only want to check that the uploads are calling the validation
- Then with unit tests we can check the types of errors raised

"""


@pytest.mark.skip()
def test_create_from_url(client, tmp_path, object_predictions,
                         model_run_with_data_rows,
                         configured_project,
                         annotation_import_test_helpers):
    name = str(uuid.uuid4())
    file_name = f"{name}.json"
    file_path = tmp_path / file_name

    model_run_data_rows = [
        mrdr.data_row().uid
        for mrdr in model_run_with_data_rows.model_run_data_rows()
    ]
    predictions = [
        p for p in object_predictions
        if p['dataRow']['id'] in model_run_data_rows
    ]
    with file_path.open("w") as f:
        parser.dump(predictions, f)

    # Needs to have data row ids

    with open(file_path, "r") as f:
        url = client.upload_data(content=f.read(),
                                 filename=file_name,
                                 sign=True,
                                 content_type="application/json")

    annotation_import, batch, mal_prediction_import = model_run_with_data_rows.upsert_predictions_and_send_to_project(
        name=name,
        predictions=url,
        project_id=configured_project.uid,
        priority=5)

    assert annotation_import.model_run_id == model_run_with_data_rows.uid
    annotation_import.wait_until_done()
    assert not annotation_import.errors
    assert annotation_import.statuses

    assert batch
    assert batch.project().uid == configured_project.uid

    assert mal_prediction_import
    mal_prediction_import.wait_until_done()

    assert not mal_prediction_import.errors
    assert mal_prediction_import.statuses


@pytest.mark.skip()
def test_create_from_objects(model_run_with_data_rows,
                             configured_project,
                             object_predictions,
                             annotation_import_test_helpers):
    name = str(uuid.uuid4())
    model_run_data_rows = [
        mrdr.data_row().uid
        for mrdr in model_run_with_data_rows.model_run_data_rows()
    ]
    predictions = [
        p for p in object_predictions
        if p['dataRow']['id'] in model_run_data_rows
    ]
    annotation_import, batch, mal_prediction_import = model_run_with_data_rows.upsert_predictions_and_send_to_project(
        name=name,
        predictions=predictions,
        project_id=configured_project.uid,
        priority=5)

    assert annotation_import.model_run_id == model_run_with_data_rows.uid
    annotation_import.wait_until_done()
    assert not annotation_import.errors
    assert annotation_import.statuses

    assert batch
    assert batch.project().uid == configured_project.uid

    assert mal_prediction_import
    mal_prediction_import.wait_until_done()

    assert not mal_prediction_import.errors
    assert mal_prediction_import.statuses


@pytest.mark.skip()
def test_create_from_local_file(tmp_path, model_run_with_data_rows,
                                configured_project_with_one_data_row,
                                object_predictions,
                                annotation_import_test_helpers):

    name = str(uuid.uuid4())
    file_name = f"{name}.ndjson"
    file_path = tmp_path / file_name

    model_run_data_rows = [
        mrdr.data_row().uid
        for mrdr in model_run_with_data_rows.model_run_data_rows()
    ]
    predictions = [
        p for p in object_predictions
        if p['dataRow']['id'] in model_run_data_rows
    ]

    with file_path.open("w") as f:
        parser.dump(predictions, f)

    annotation_import, batch, mal_prediction_import = model_run_with_data_rows.upsert_predictions_and_send_to_project(
        name=name,
        predictions=str(file_path),
        project_id=configured_project_with_one_data_row.uid,
        priority=5)

    assert annotation_import.model_run_id == model_run_with_data_rows.uid
    annotation_import.wait_until_done()
    assert not annotation_import.errors
    assert annotation_import.statuses

    assert batch
    assert batch.project().uid == configured_project_with_one_data_row.uid

    assert mal_prediction_import
    mal_prediction_import.wait_until_done()

    assert not mal_prediction_import.errors
    assert mal_prediction_import.statuses



from types import SimpleNamespace

import pytest

from labelbox.data.annotation_types import ClassificationAnnotation, ObjectAnnotation
from labelbox.data.annotation_types import Polygon, Point, Rectangle, Mask, MaskData, Line, Radio, Text, Checklist, ClassificationAnswer
import numpy as np

from labelbox.data.annotation_types.ner import TextEntity


class NameSpace(SimpleNamespace):

    def __init__(self,
                 predictions,
                 ground_truths,
                 expected,
                 expected_without_subclasses=None):
        super(NameSpace, self).__init__(
            predictions=predictions,
            ground_truths=ground_truths,
            expected=expected,
            expected_without_subclasses=expected_without_subclasses or expected)


def get_radio(name, answer_name):
    return ClassificationAnnotation(
        name=name, value=Radio(answer=ClassificationAnswer(name=answer_name)))


def get_text(name, text_content):
    return ClassificationAnnotation(name=name, value=Text(answer=text_content))


def get_checklist(name, answer_names):
    return ClassificationAnnotation(name=name,
                                    value=Radio(answer=[
                                        ClassificationAnswer(name=answer_name)
                                        for answer_name in answer_names
                                    ]))


def get_polygon(name, points, subclasses=None):
    return ObjectAnnotation(
        name=name,
        value=Polygon(points=[Point(x=x, y=y) for x, y in points]),
        classifications=[] if subclasses is None else subclasses)


def get_rectangle(name, start, end, subclasses=None):
    return ObjectAnnotation(
        name=name,
        value=Rectangle(start=Point(x=start[0], y=start[1]),
                        end=Point(x=end[0], y=end[1])),
        classifications=[] if subclasses is None else subclasses)


def get_mask(name, pixels, color=(1, 1, 1), subclasses=None):
    mask = np.zeros((32, 32, 3)).astype(np.uint8)
    for pixel in pixels:
        mask[pixel[0], pixel[1]] = color
    return ObjectAnnotation(
        name=name,
        value=Mask(mask=MaskData(arr=mask), color=color),
        classifications=[] if subclasses is None else subclasses)


def get_line(name, points, subclasses=None):
    return ObjectAnnotation(
        name=name,
        value=Line(points=[Point(x=x, y=y) for x, y in points]),
        classifications=[] if subclasses is None else subclasses)


def get_point(name, x, y, subclasses=None):
    return ObjectAnnotation(
        name=name,
        value=Point(x=x, y=y),
        classifications=[] if subclasses is None else subclasses)


def get_radio(name, answer_name):
    return ClassificationAnnotation(
        name=name, value=Radio(answer=ClassificationAnswer(name=answer_name)))


def get_checklist(name, answer_names):
    return ClassificationAnnotation(name=name,
                                    value=Checklist(answer=[
                                        ClassificationAnswer(name=answer_name)
                                        for answer_name in answer_names
                                    ]))


def get_ner(name, start, end, subclasses=None):
    return ObjectAnnotation(
        name=name,
        value=TextEntity(start=start, end=end),
        classifications=[] if subclasses is None else subclasses)


def get_object_pairs(tool_fn, **kwargs):
    return [
        NameSpace(predictions=[tool_fn("cat", **kwargs)],
                  ground_truths=[tool_fn("cat", **kwargs)],
                  expected={'cat': [1, 0, 0, 0]}),
        NameSpace(
            predictions=[
                tool_fn("cat",
                        **kwargs,
                        subclasses=[get_radio("is_animal", answer_name="yes")])
            ],
            ground_truths=[
                tool_fn("cat",
                        **kwargs,
                        subclasses=[get_radio("is_animal", answer_name="yes")])
            ],
            expected={'cat': [1, 0, 0, 0]},
            expected_without_subclasses={'cat': [1, 0, 0, 0]}),
        NameSpace(predictions=[
            tool_fn("cat",
                    **kwargs,
                    subclasses=[get_radio("is_animal", answer_name="yes")])
        ],
                  ground_truths=[
                      tool_fn(
                          "cat",
                          **kwargs,
                          subclasses=[get_radio("is_animal", answer_name="no")])
                  ],
                  expected={'cat': [0, 1, 0, 1]},
                  expected_without_subclasses={'cat': [1, 0, 0, 0]}),
        NameSpace(predictions=[
            tool_fn("cat",
                    **kwargs,
                    subclasses=[get_radio("is_animal", answer_name="yes")]),
            tool_fn("cat",
                    **kwargs,
                    subclasses=[get_radio("is_animal", answer_name="no")])
        ],
                  ground_truths=[
                      tool_fn(
                          "cat",
                          **kwargs,
                          subclasses=[get_radio("is_animal", answer_name="no")])
                  ],
                  expected={'cat': [1, 1, 0, 0]},
                  expected_without_subclasses={'cat': [1, 1, 0, 0]}),
        NameSpace(predictions=[
            tool_fn("cat",
                    **kwargs,
                    subclasses=[get_radio("is_animal", answer_name="yes")]),
            tool_fn("dog",
                    **kwargs,
                    subclasses=[get_radio("is_animal", answer_name="no")])
        ],
                  ground_truths=[
                      tool_fn(
                          "cat",
                          **kwargs,
                          subclasses=[get_radio("is_animal", answer_name="no")])
                  ],
                  expected={
                      'cat': [0, 1, 0, 1],
                      'dog': [0, 1, 0, 0]
                  },
                  expected_without_subclasses={
                      'cat': [1, 0, 0, 0],
                      'dog': [0, 1, 0, 0]
                  }),
        NameSpace(
            predictions=[tool_fn("cat", **kwargs),
                         tool_fn("cat", **kwargs)],
            ground_truths=[tool_fn("cat", **kwargs),
                           tool_fn("cat", **kwargs)],
            expected={'cat': [2, 0, 0, 0]}),
        NameSpace(
            predictions=[tool_fn("cat", **kwargs),
                         tool_fn("cat", **kwargs)],
            ground_truths=[tool_fn("cat", **kwargs)],
            expected={'cat': [1, 1, 0, 0]}),
        NameSpace(
            predictions=[tool_fn("cat", **kwargs)],
            ground_truths=[tool_fn("cat", **kwargs),
                           tool_fn("cat", **kwargs)],
            expected={'cat': [1, 0, 0, 1]}),
        NameSpace(predictions=[],
                  ground_truths=[],
                  expected=[],
                  expected_without_subclasses=[]),
        NameSpace(predictions=[],
                  ground_truths=[tool_fn("cat", **kwargs)],
                  expected={'cat': [0, 0, 0, 1]}),
        NameSpace(predictions=[tool_fn("cat", **kwargs)],
                  ground_truths=[],
                  expected={'cat': [0, 1, 0, 0]}),
        NameSpace(predictions=[tool_fn("cat", **kwargs)],
                  ground_truths=[tool_fn("dog", **kwargs)],
                  expected={
                      'cat': [0, 1, 0, 0],
                      'dog': [0, 0, 0, 1]
                  })
    ]


@pytest.fixture
def radio_pairs():
    return [
        NameSpace(predictions=[get_radio("is_animal", answer_name="yes")],
                  ground_truths=[get_radio("is_animal", answer_name="yes")],
                  expected={'yes': [1, 0, 0, 0]}),
        NameSpace(predictions=[get_radio("is_animal", answer_name="yes")],
                  ground_truths=[get_radio("is_animal", answer_name="no")],
                  expected={
                      'no': [0, 0, 0, 1],
                      'yes': [0, 1, 0, 0]
                  }),
        NameSpace(predictions=[get_radio("is_animal", answer_name="yes")],
                  ground_truths=[],
                  expected={'yes': [0, 1, 0, 0]}),
        NameSpace(predictions=[],
                  ground_truths=[get_radio("is_animal", answer_name="yes")],
                  expected={'yes': [0, 0, 0, 1]}),
        NameSpace(predictions=[
            get_radio("is_animal", answer_name="yes"),
            get_radio("is_short", answer_name="no")
        ],
                  ground_truths=[get_radio("is_animal", answer_name="yes")],
                  expected={
                      'no': [0, 1, 0, 0],
                      'yes': [1, 0, 0, 0]
                  }),
        #Not supported yet:
        # NameSpace(
        #predictions=[],
        #ground_truths=[],
        #expected = [0,0,1,0]
        #)
    ]


@pytest.fixture
def checklist_pairs():
    return [
        NameSpace(predictions=[
            get_checklist("animal_attributes", answer_names=["striped"])
        ],
                  ground_truths=[
                      get_checklist("animal_attributes",
                                    answer_names=["striped"])
                  ],
                  expected={'striped': [1, 0, 0, 0]}),
        NameSpace(predictions=[
            get_checklist("animal_attributes", answer_names=["striped"])
        ],
                  ground_truths=[],
                  expected={'striped': [0, 1, 0, 0]}),
        NameSpace(predictions=[],
                  ground_truths=[
                      get_checklist("animal_attributes",
                                    answer_names=["striped"])
                  ],
                  expected={'striped': [0, 0, 0, 1]}),
        NameSpace(predictions=[
            get_checklist("animal_attributes",
                          answer_names=["striped", "short"])
        ],
                  ground_truths=[
                      get_checklist("animal_attributes",
                                    answer_names=["striped"])
                  ],
                  expected={
                      'short': [0, 1, 0, 0],
                      'striped': [1, 0, 0, 0]
                  }),
        NameSpace(predictions=[
            get_checklist("animal_attributes", answer_names=["striped"])
        ],
                  ground_truths=[
                      get_checklist("animal_attributes",
                                    answer_names=["striped", "short"])
                  ],
                  expected={
                      'short': [0, 0, 0, 1],
                      'striped': [1, 0, 0, 0]
                  }),
        NameSpace(predictions=[
            get_checklist("animal_attributes",
                          answer_names=["striped", "short", "black"])
        ],
                  ground_truths=[
                      get_checklist("animal_attributes",
                                    answer_names=["striped", "short"])
                  ],
                  expected={
                      'black': [0, 1, 0, 0],
                      'short': [1, 0, 0, 0],
                      'striped': [1, 0, 0, 0]
                  }),
        NameSpace(predictions=[
            get_checklist("animal_attributes",
                          answer_names=["striped", "short", "black"]),
            get_checklist("animal_name", answer_names=["doggy", "pup"])
        ],
                  ground_truths=[
                      get_checklist("animal_attributes",
                                    answer_names=["striped", "short"]),
                      get_checklist("animal_name", answer_names=["pup"])
                  ],
                  expected={
                      'black': [0, 1, 0, 0],
                      'doggy': [0, 1, 0, 0],
                      'pup': [1, 0, 0, 0],
                      'short': [1, 0, 0, 0],
                      'striped': [1, 0, 0, 0]
                  })

        #Not supported yet:
        # NameSpace(
        #predictions=[],
        #ground_truths=[],
        #expected = [0,0,1,0]
        #)
    ]


@pytest.fixture
def polygon_pairs():
    return get_object_pairs(get_polygon,
                            points=[[0, 0], [10, 0], [10, 10], [0, 10]])


@pytest.fixture
def rectangle_pairs():
    return get_object_pairs(get_rectangle, start=[0, 0], end=[10, 10])


@pytest.fixture
def mask_pairs():
    return get_object_pairs(get_mask, pixels=[[0, 0]])


@pytest.fixture
def line_pairs():
    return get_object_pairs(get_line,
                            points=[[0, 0], [10, 0], [10, 10], [0, 10]])


@pytest.fixture
def point_pairs():
    return get_object_pairs(get_point, x=0, y=0)


@pytest.fixture
def ner_pairs():
    return get_object_pairs(get_ner, start=0, end=10)


@pytest.fixture()
def pair_iou_thresholds():
    return [
        NameSpace(predictions=[
            get_polygon("cat", points=[[0, 0], [10, 0], [10, 10], [0, 10]]),
        ],
                  ground_truths=[
                      get_polygon("cat",
                                  points=[[0, 0], [5, 0], [5, 5], [0, 5]]),
                  ],
                  expected={
                      0.2: [1, 0, 0, 0],
                      0.3: [0, 1, 0, 1]
                  }),
        NameSpace(
            predictions=[get_rectangle("cat", start=[0, 0], end=[10, 10])],
            ground_truths=[get_rectangle("cat", start=[0, 0], end=[5, 5])],
            expected={
                0.2: [1, 0, 0, 0],
                0.3: [0, 1, 0, 1]
            }),
        NameSpace(predictions=[get_point("cat", x=0, y=0)],
                  ground_truths=[get_point("cat", x=20, y=20)],
                  expected={
                      0.5: [1, 0, 0, 0],
                      0.65: [0, 1, 0, 1]
                  }),
        NameSpace(predictions=[
            get_line("cat", points=[[0, 0], [10, 0], [10, 10], [0, 10]])
        ],
                  ground_truths=[
                      get_line("cat",
                               points=[[0, 0], [100, 0], [100, 100], [0, 100]])
                  ],
                  expected={
                      0.3: [1, 0, 0, 0],
                      0.65: [0, 1, 0, 1]
                  }),
        NameSpace(predictions=[
            get_mask("cat", pixels=[[0, 0], [1, 1], [2, 2], [3, 3]])
        ],
                  ground_truths=[get_mask("cat", pixels=[[0, 0], [1, 1]])],
                  expected={
                      0.4: [1, 0, 0, 0],
                      0.6: [0, 1, 0, 1]
                  }),
    ]



from pathlib import Path

from typing import Optional, Tuple
from PIL import Image
import imagesize

from .path import PathSerializerMixin
from ...annotation_types import Label


class CocoImage(PathSerializerMixin):
    id: int
    width: int
    height: int
    file_name: Path
    license: Optional[int] = None
    flickr_url: Optional[str] = None
    coco_url: Optional[str] = None


def get_image_id(label: Label, idx: int) -> int:
    if label.data.file_path is not None:
        file_name = label.data.file_path.replace(".jpg", "")
        if file_name.isdecimal():
            return file_name
    return idx


def get_image(label: Label, image_root: Path, image_id: str) -> CocoImage:
    path = Path(image_root, f"{image_id}.jpg")
    if not path.exists():
        im = Image.fromarray(label.data.value)
        im.save(path)
        w, h = im.size
    else:
        w, h = imagesize.get(str(path))
    return CocoImage(id=image_id, width=w, height=h, file_name=Path(path.name))


def id_to_rgb(id: int) -> Tuple[int, int, int]:
    digits = []
    for _ in range(3):
        digits.append(id % 256)
        id //= 256
    return digits


def rgb_to_id(red: int, green: int, blue: int) -> int:
    id = blue * 256 * 256
    id += (green * 256)
    id += red
    return id



from concurrent.futures import ProcessPoolExecutor, as_completed
from typing import Dict, Any, List, Union
from pathlib import Path

from labelbox import pydantic_compat
from tqdm import tqdm
import numpy as np
from PIL import Image

from ...annotation_types.geometry import Polygon, Rectangle
from ...annotation_types import Label
from ...annotation_types.geometry.mask import Mask
from ...annotation_types.annotation import ObjectAnnotation
from ...annotation_types.data.raster import MaskData, ImageData
from ...annotation_types.collection import LabelCollection
from .categories import Categories, hash_category_name
from .image import CocoImage, get_image, get_image_id, id_to_rgb
from .annotation import PanopticAnnotation, SegmentInfo, get_annotation_lookup


def vector_to_coco_segment_info(canvas: np.ndarray,
                                annotation: ObjectAnnotation,
                                annotation_idx: int, image: CocoImage,
                                category_id: int):

    shapely = annotation.value.shapely
    if shapely.is_empty:
        return

    xmin, ymin, xmax, ymax = shapely.bounds
    canvas = annotation.value.draw(height=image.height,
                                   width=image.width,
                                   canvas=canvas,
                                   color=id_to_rgb(annotation_idx))

    return SegmentInfo(id=annotation_idx,
                       category_id=category_id,
                       area=shapely.area,
                       bbox=[xmin, ymin, xmax - xmin, ymax - ymin]), canvas


def mask_to_coco_segment_info(canvas: np.ndarray, annotation,
                              annotation_idx: int, category_id):
    color = id_to_rgb(annotation_idx)
    mask = annotation.value.draw(color=color)
    shapely = annotation.value.shapely
    if shapely.is_empty:
        return

    xmin, ymin, xmax, ymax = shapely.bounds
    canvas = np.where(canvas == (0, 0, 0), mask, canvas)
    return SegmentInfo(id=annotation_idx,
                       category_id=category_id,
                       area=shapely.area,
                       bbox=[xmin, ymin, xmax - xmin, ymax - ymin]), canvas


def process_label(label: Label,
                  idx: Union[int, str],
                  image_root,
                  mask_root,
                  all_stuff=False):
    """
    Masks become stuff
    Polygon and rectangle become thing
    """
    annotations = get_annotation_lookup(label.annotations)
    image_id = get_image_id(label, idx)
    image = get_image(label, image_root, image_id)
    canvas = np.zeros((image.height, image.width, 3))

    segments = []
    categories = {}
    is_thing = {}

    for class_idx, class_name in enumerate(annotations):
        for annotation_idx, annotation in enumerate(annotations[class_name]):
            categories[annotation.name] = hash_category_name(annotation.name)
            if isinstance(annotation.value, Mask):
                coco_segment_info = mask_to_coco_segment_info(
                    canvas, annotation, class_idx + 1,
                    categories[annotation.name])

                if coco_segment_info is None:
                    # Filter out empty masks
                    continue

                segment, canvas = coco_segment_info
                segments.append(segment)
                is_thing[annotation.name] = 0

            elif isinstance(annotation.value, (Polygon, Rectangle)):
                coco_vector_info = vector_to_coco_segment_info(
                    canvas,
                    annotation,
                    annotation_idx=(class_idx if all_stuff else annotation_idx)
                    + 1,
                    image=image,
                    category_id=categories[annotation.name])

                if coco_vector_info is None:
                    # Filter out empty annotations
                    continue

                segment, canvas = coco_vector_info
                segments.append(segment)
                is_thing[annotation.name] = 1 - int(all_stuff)

    mask_file = str(image.file_name).replace('.jpg', '.png')
    mask_file = Path(mask_root, mask_file)
    Image.fromarray(canvas.astype(np.uint8)).save(mask_file)
    return image, PanopticAnnotation(
        image_id=image_id,
        file_name=Path(mask_file.name),
        segments_info=segments), categories, is_thing


class CocoPanopticDataset(pydantic_compat.BaseModel):
    info: Dict[str, Any] = {}
    images: List[CocoImage]
    annotations: List[PanopticAnnotation]
    categories: List[Categories]

    @classmethod
    def from_common(cls,
                    labels: LabelCollection,
                    image_root,
                    mask_root,
                    all_stuff,
                    max_workers=8):
        all_coco_annotations = []
        coco_categories = {}
        coco_things = {}
        images = []

        if max_workers:
            with ProcessPoolExecutor(max_workers=max_workers) as exc:
                futures = [
                    exc.submit(process_label, label, idx, image_root, mask_root,
                               all_stuff) for idx, label in enumerate(labels)
                ]
                results = [
                    future.result() for future in tqdm(as_completed(futures))
                ]
        else:
            results = [
                process_label(label, idx, image_root, mask_root, all_stuff)
                for idx, label in enumerate(labels)
            ]

        for result in results:
            images.append(result[0])
            all_coco_annotations.append(result[1])
            coco_categories.update(result[2])
            coco_things.update(result[3])

        category_mapping = {
            category_id: idx + 1
            for idx, category_id in enumerate(coco_categories.values())
        }
        categories = [
            Categories(id=category_mapping[idx],
                       name=name,
                       supercategory='all',
                       isthing=coco_things.get(name, 1))
            for name, idx in coco_categories.items()
        ]

        for annot in all_coco_annotations:
            for segment in annot.segments_info:
                segment.category_id = category_mapping[segment.category_id]

        return CocoPanopticDataset(info={
            'image_root': image_root,
            'mask_root': mask_root
        },
                                   images=images,
                                   annotations=all_coco_annotations,
                                   categories=categories)

    def to_common(self, image_root: Path, mask_root: Path):
        category_lookup = {
            category.id: category for category in self.categories
        }
        annotation_lookup = {
            annotation.image_id: annotation for annotation in self.annotations
        }
        for image in self.images:
            annotations = []
            annotation = annotation_lookup[image.id]

            im_path = Path(image_root, image.file_name)
            if not im_path.exists():
                raise ValueError(
                    f"Cannot find file {im_path}. Make sure `image_root` is set properly"
                )
            if not str(annotation.file_name).endswith('.png'):
                raise ValueError(
                    f"COCO masks must be stored as png files and their extension must be `.png`. Found {annotation.file_name}"
                )
            mask = MaskData(
                file_path=str(Path(mask_root, annotation.file_name)))

            for segmentation in annotation.segments_info:
                category = category_lookup[segmentation.category_id]
                annotations.append(
                    ObjectAnnotation(name=category.name,
                                     value=Mask(mask=mask,
                                                color=id_to_rgb(
                                                    segmentation.id))))
            data = ImageData(file_path=str(im_path))
            yield Label(data=data, annotations=annotations)
            del annotation_lookup[image.id]



from typing import Dict, Any, Union
from pathlib import Path
import os
import warnings

from ...annotation_types.collection import LabelCollection, LabelGenerator
from ...serialization.coco.instance_dataset import CocoInstanceDataset
from ...serialization.coco.panoptic_dataset import CocoPanopticDataset


def create_path_if_not_exists(path: Union[Path, str],
                              ignore_existing_data=False):
    path = Path(path)
    if not path.exists():
        path.mkdir(parents=True, exist_ok=True)
    elif not ignore_existing_data and os.listdir(path):
        raise ValueError(
            f"Directory `{path}`` must be empty. Or set `ignore_existing_data=True`"
        )
    return path


def validate_path(path: Union[Path, str], name: str):
    path = Path(path)
    if not path.exists():
        raise ValueError(f"{name} `{path}` must exist")
    return path


class COCOConverter:
    """
    Class for converting between coco and labelbox formats
    Note that this class is only compatible with image data.

    Subclasses are currently ignored.
    To use subclasses, manually flatten them before using the converter.
    """

    @staticmethod
    def serialize_instances(labels: LabelCollection,
                            image_root: Union[Path, str],
                            ignore_existing_data=False,
                            max_workers=8) -> Dict[str, Any]:
        """
        Convert a Labelbox LabelCollection into an mscoco dataset.
        This function will only convert masks, polygons, and rectangles.
        Masks will be converted into individual instances.
        Use deserialize_panoptic to prevent masks from being split apart.

        Args:
            labels: A collection of labels to convert
            image_root: Where to save images to
            ignore_existing_data: Whether or not to raise an exception if images already exist.
                This exists only to support detectons panoptic fpn model which requires two mscoco payloads for the same images.
            max_workers : Number of workers to process dataset with. A value of 0 will process all data in the main process
        Returns:
            A dictionary containing labels in the coco object format.
        """

        warnings.warn(
            "You are currently utilizing COCOconverter for this action, which will be deprecated in a later release.",
            DeprecationWarning,
            stacklevel=2)

        image_root = create_path_if_not_exists(image_root, ignore_existing_data)
        return CocoInstanceDataset.from_common(labels=labels,
                                               image_root=image_root,
                                               max_workers=max_workers).dict()

    @staticmethod
    def serialize_panoptic(labels: LabelCollection,
                           image_root: Union[Path, str],
                           mask_root: Union[Path, str],
                           all_stuff: bool = False,
                           ignore_existing_data=False,
                           max_workers: int = 8) -> Dict[str, Any]:
        """
        Convert a Labelbox LabelCollection into an mscoco dataset.
        This function will only convert masks, polygons, and rectangles.
        Masks will be converted into individual instances.
        Use deserialize_panoptic to prevent masks from being split apart.

        Args:
            labels: A collection of labels to convert
            image_root: Where to save images to
            mask_root: Where to save segmentation masks to
            all_stuff: If rectangle or polygon annotations are encountered, they will be treated as instances.
                To convert them to stuff class set `all_stuff=True`.
            ignore_existing_data: Whether or not to raise an exception if images already exist.
                This exists only to support detectons panoptic fpn model which requires two mscoco payloads for the same images.
            max_workers : Number of workers to process dataset with. A value of 0 will process all data in the main process.
        Returns:
            A dictionary containing labels in the coco panoptic format.
        """

        warnings.warn(
            "You are currently utilizing COCOconverter for this action, which will be deprecated in a later release.",
            DeprecationWarning,
            stacklevel=2)

        image_root = create_path_if_not_exists(image_root, ignore_existing_data)
        mask_root = create_path_if_not_exists(mask_root, ignore_existing_data)
        return CocoPanopticDataset.from_common(labels=labels,
                                               image_root=image_root,
                                               mask_root=mask_root,
                                               all_stuff=all_stuff,
                                               max_workers=max_workers).dict()

    @staticmethod
    def deserialize_panoptic(json_data: Dict[str, Any], image_root: Union[Path,
                                                                          str],
                             mask_root: Union[Path, str]) -> LabelGenerator:
        """
        Convert coco panoptic data into the labelbox format (as a LabelGenerator).

        Args:
            json_data: panoptic data as a dict
            image_root: Path to local images that are referenced by the panoptic json
            mask_root: Path to local segmentation masks that are referenced by the panoptic json
        Returns:
            LabelGenerator
        """

        warnings.warn(
            "You are currently utilizing COCOconverter for this action, which will be deprecated in a later release.",
            DeprecationWarning,
            stacklevel=2)

        image_root = validate_path(image_root, 'image_root')
        mask_root = validate_path(mask_root, 'mask_root')
        objs = CocoPanopticDataset(**json_data)
        gen = objs.to_common(image_root, mask_root)
        return LabelGenerator(data=gen)

    @staticmethod
    def deserialize_instances(json_data: Dict[str, Any],
                              image_root: Path) -> LabelGenerator:
        """
        Convert coco object data into the labelbox format (as a LabelGenerator).

        Args:
            json_data: coco object data as a dict
            image_root: Path to local images that are referenced by the coco object json
        Returns:
            LabelGenerator
        """

        warnings.warn(
            "You are currently utilizing COCOconverter for this action, which will be deprecated in a later release.",
            DeprecationWarning,
            stacklevel=2)

        image_root = validate_path(image_root, 'image_root')
        objs = CocoInstanceDataset(**json_data)
        gen = objs.to_common(image_root)
        return LabelGenerator(data=gen)



import json

import pytest
import labelbox.types as lb_types
from labelbox.data.serialization.ndjson.converter import NDJsonConverter

radio_ndjson = [{
    'dataRow': {
        'globalKey': 'my_global_key'
    },
    'name': 'radio',
    'answer': {
        'name': 'first_radio_answer'
    },
    'messageId': '0'
}]

radio_label = [
    lb_types.Label(
        data=lb_types.ConversationData(global_key='my_global_key'),
        annotations=[
            lb_types.ClassificationAnnotation(
                name='radio',
                value=lb_types.Radio(answer=lb_types.ClassificationAnswer(
                    name="first_radio_answer")),
                message_id="0")
        ])
]

checklist_ndjson = [{
    'dataRow': {
        'globalKey': 'my_global_key'
    },
    'name': 'checklist',
    'answer': [
        {
            'name': 'first_checklist_answer'
        },
        {
            'name': 'second_checklist_answer'
        },
    ],
    'messageId': '2'
}]

checklist_label = [
    lb_types.Label(data=lb_types.ConversationData(global_key='my_global_key'),
                   annotations=[
                       lb_types.ClassificationAnnotation(
                           name='checklist',
                           message_id="2",
                           value=lb_types.Checklist(answer=[
                               lb_types.ClassificationAnswer(
                                   name="first_checklist_answer"),
                               lb_types.ClassificationAnswer(
                                   name="second_checklist_answer")
                           ]))
                   ])
]

free_text_ndjson = [{
    'dataRow': {
        'globalKey': 'my_global_key'
    },
    'name': 'free_text',
    'answer': 'sample text',
    'messageId': '0'
}]
free_text_label = [
    lb_types.Label(data=lb_types.ConversationData(global_key='my_global_key'),
                   annotations=[
                       lb_types.ClassificationAnnotation(
                           name='free_text',
                           message_id="0",
                           value=lb_types.Text(answer="sample text"))
                   ])
]


@pytest.mark.parametrize(
    "label, ndjson",
    [[radio_label, radio_ndjson], [checklist_label, checklist_ndjson],
     [free_text_label, free_text_ndjson]])
def test_message_based_radio_classification(label, ndjson):
    serialized_label = list(NDJsonConverter().serialize(label))
    serialized_label[0].pop('uuid')
    assert serialized_label == ndjson

    deserialized_label = list(NDJsonConverter().deserialize(ndjson))
    deserialized_label[0].annotations[0].extra.pop('uuid')
    assert deserialized_label[0].annotations == label[0].annotations


@pytest.mark.parametrize("filename", [
    "tests/data/assets/ndjson/conversation_entity_import.json",
    "tests/data/assets/ndjson/conversation_entity_without_confidence_import.json"
])
def test_conversation_entity_import(filename: str):
    with open(filename, 'r') as file:
        data = json.load(file)
    res = list(NDJsonConverter.deserialize(data))
    res = list(NDJsonConverter.serialize(res))
    assert res == data



import pytest

from labelbox import OntologyBuilder, MediaType, Tool
from labelbox.orm.model import Entity
import json
import time

from labelbox.schema.queue_mode import QueueMode


def test_feature_schema_is_not_archived(client, ontology):
    feature_schema_to_check = ontology.normalized['tools'][0]
    result = client.is_feature_schema_archived(
        ontology.uid, feature_schema_to_check['featureSchemaId'])
    assert result == False


def test_feature_schema_is_archived(client, configured_project_with_label):
    project, _, _, label = configured_project_with_label
    ontology = project.ontology()
    feature_schema_id = ontology.normalized['tools'][0]['featureSchemaId']
    result = client.delete_feature_schema_from_ontology(ontology.uid,
                                                        feature_schema_id)
    assert result.archived == True and result.deleted == False
    assert client.is_feature_schema_archived(ontology.uid,
                                             feature_schema_id) == True


def test_is_feature_schema_archived_for_non_existing_feature_schema(
        client, ontology):
    with pytest.raises(
            Exception,
            match="The specified feature schema was not in the ontology"):
        client.is_feature_schema_archived(ontology.uid,
                                          'invalid-feature-schema-id')


def test_is_feature_schema_archived_for_non_existing_ontology(client, ontology):
    feature_schema_to_unarchive = ontology.normalized['tools'][0]
    with pytest.raises(
            Exception,
            match="Resource 'Ontology' not found for params: 'invalid-ontology'"
    ):
        client.is_feature_schema_archived(
            'invalid-ontology', feature_schema_to_unarchive['featureSchemaId'])


def test_delete_tool_feature_from_ontology(client, ontology):
    feature_schema_to_delete = ontology.normalized['tools'][0]
    assert len(ontology.normalized['tools']) == 2
    result = client.delete_feature_schema_from_ontology(
        ontology.uid, feature_schema_to_delete['featureSchemaId'])
    assert result.deleted == True
    assert result.archived == False
    updatedOntology = client.get_ontology(ontology.uid)
    assert len(updatedOntology.normalized['tools']) == 1


@pytest.mark.skip(reason="normalized ontology contains Relationship, "
                  "which is not finalized yet. introduce this back when"
                  "Relationship feature is complete and we introduce"
                  "a Relationship object to the ontology that we can parse")
def test_from_project_ontology(project) -> None:
    o = OntologyBuilder.from_project(project)
    assert o.asdict() == project.ontology().normalized


point = Tool(
    tool=Tool.Type.POINT,
    name="name",
    color="#ff0000",
)


def test_deletes_an_ontology(client):
    tool = client.upsert_feature_schema(point.asdict())
    feature_schema_id = tool.normalized['featureSchemaId']
    ontology = client.create_ontology_from_feature_schemas(
        name='ontology name',
        feature_schema_ids=[feature_schema_id],
        media_type=MediaType.Image)

    assert client.delete_unused_ontology(ontology.uid) is None

    client.delete_unused_feature_schema(feature_schema_id)


def test_cant_delete_an_ontology_with_project(client):
    project = client.create_project(name="test project",
                                    queue_mode=QueueMode.Batch,
                                    media_type=MediaType.Image)
    tool = client.upsert_feature_schema(point.asdict())
    feature_schema_id = tool.normalized['featureSchemaId']
    ontology = client.create_ontology_from_feature_schemas(
        name='ontology name',
        feature_schema_ids=[feature_schema_id],
        media_type=MediaType.Image)
    project.connect_ontology(ontology)

    with pytest.raises(
            Exception,
            match=
            "Failed to delete the ontology, message: Cannot delete an ontology connected to a project. The ontology is connected to projects: "
            + project.uid):
        client.delete_unused_ontology(ontology.uid)

    project.delete()
    client.delete_unused_ontology(ontology.uid)
    client.delete_unused_feature_schema(feature_schema_id)


def test_inserts_a_feature_schema_at_given_position(client):
    tool1 = {'tool': 'polygon', 'name': 'tool1', 'color': 'blue'}
    tool2 = {'tool': 'polygon', 'name': 'tool2', 'color': 'blue'}
    ontology_normalized_json = {"tools": [tool1, tool2], "classifications": []}
    ontology = client.create_ontology(name="ontology",
                                      normalized=ontology_normalized_json,
                                      media_type=MediaType.Image)
    created_feature_schema = client.upsert_feature_schema(point.asdict())
    client.insert_feature_schema_into_ontology(
        created_feature_schema.normalized['featureSchemaId'], ontology.uid, 1)
    ontology = client.get_ontology(ontology.uid)

    assert ontology.normalized['tools'][1][
        'schemaNodeId'] == created_feature_schema.normalized['schemaNodeId']

    client.delete_unused_ontology(ontology.uid)


def test_moves_already_added_feature_schema_in_ontology(client):
    tool1 = {'tool': 'polygon', 'name': 'tool1', 'color': 'blue'}
    ontology_normalized_json = {"tools": [tool1], "classifications": []}
    ontology = client.create_ontology(name="ontology",
                                      normalized=ontology_normalized_json,
                                      media_type=MediaType.Image)
    created_feature_schema = client.upsert_feature_schema(point.asdict())
    feature_schema_id = created_feature_schema.normalized['featureSchemaId']
    client.insert_feature_schema_into_ontology(feature_schema_id, ontology.uid,
                                               1)
    ontology = client.get_ontology(ontology.uid)
    assert ontology.normalized['tools'][1][
        'schemaNodeId'] == created_feature_schema.normalized['schemaNodeId']
    client.insert_feature_schema_into_ontology(feature_schema_id, ontology.uid,
                                               0)
    ontology = client.get_ontology(ontology.uid)

    assert ontology.normalized['tools'][0][
        'schemaNodeId'] == created_feature_schema.normalized['schemaNodeId']

    client.delete_unused_ontology(ontology.uid)


def test_does_not_include_used_ontologies(client):
    tool = client.upsert_feature_schema(point.asdict())
    feature_schema_id = tool.normalized['featureSchemaId']
    ontology_with_project = client.create_ontology_from_feature_schemas(
        name='ontology name',
        feature_schema_ids=[feature_schema_id],
        media_type=MediaType.Image)
    project = client.create_project(name="test project",
                                    queue_mode=QueueMode.Batch,
                                    media_type=MediaType.Image)
    project.connect_ontology(ontology_with_project)
    unused_ontologies = client.get_unused_ontologies()

    assert ontology_with_project.uid not in unused_ontologies

    project.delete()
    client.delete_unused_ontology(ontology_with_project.uid)
    client.delete_unused_feature_schema(feature_schema_id)


def _get_attr_stringify_json(obj, attr):
    value = getattr(obj, attr.name)
    if attr.field_type.name.lower() == "json":
        return json.dumps(value, sort_keys=True)
    return value


def test_feature_schema_create_read(client, rand_gen):
    name = f"test-root-schema-{rand_gen(str)}"
    feature_schema_cat_normalized = {
        'tool': 'polygon',
        'name': name,
        'color': 'black',
        'classifications': [],
    }
    created_feature_schema = client.create_feature_schema(
        feature_schema_cat_normalized)
    queried_feature_schema = client.get_feature_schema(
        created_feature_schema.uid)
    for attr in Entity.FeatureSchema.fields():
        assert _get_attr_stringify_json(created_feature_schema,
                                        attr) == _get_attr_stringify_json(
                                            queried_feature_schema, attr)

    time.sleep(3)  # Slight delay for searching
    queried_feature_schemas = list(client.get_feature_schemas(name))
    assert [feature_schema.name for feature_schema in queried_feature_schemas
           ] == [name]
    queried_feature_schema = queried_feature_schemas[0]

    for attr in Entity.FeatureSchema.fields():
        assert _get_attr_stringify_json(created_feature_schema,
                                        attr) == _get_attr_stringify_json(
                                            queried_feature_schema, attr)


def test_ontology_create_read(client, rand_gen):
    ontology_name = f"test-ontology-{rand_gen(str)}"
    tool_name = f"test-ontology-tool-{rand_gen(str)}"
    feature_schema_cat_normalized = {
        'tool': 'polygon',
        'name': tool_name,
        'color': 'black',
        'classifications': [],
    }
    feature_schema = client.create_feature_schema(feature_schema_cat_normalized)
    created_ontology = client.create_ontology_from_feature_schemas(
        name=ontology_name,
        feature_schema_ids=[feature_schema.uid],
        media_type=MediaType.Image)
    tool_normalized = created_ontology.normalized['tools'][0]
    for k, v in feature_schema_cat_normalized.items():
        assert tool_normalized[k] == v
    assert tool_normalized['schemaNodeId'] is not None
    assert tool_normalized['featureSchemaId'] == feature_schema.uid

    queried_ontology = client.get_ontology(created_ontology.uid)

    for attr in Entity.Ontology.fields():
        assert _get_attr_stringify_json(created_ontology,
                                        attr) == _get_attr_stringify_json(
                                            queried_ontology, attr)

    time.sleep(3)  # Slight delay for searching
    queried_ontologies = list(client.get_ontologies(ontology_name))
    assert [ontology.name for ontology in queried_ontologies] == [ontology_name]
    queried_ontology = queried_ontologies[0]
    for attr in Entity.Ontology.fields():
        assert _get_attr_stringify_json(created_ontology,
                                        attr) == _get_attr_stringify_json(
                                            queried_ontology, attr)


def test_unarchive_feature_schema_node(client, ontology):
    feature_schema_to_unarchive = ontology.normalized['tools'][0]
    result = client.unarchive_feature_schema_node(
        ontology.uid, feature_schema_to_unarchive['featureSchemaId'])
    assert result == None


def test_unarchive_feature_schema_node_for_non_existing_feature_schema(
        client, ontology):
    with pytest.raises(
            Exception,
            match=
            "Failed to find feature schema node by id: invalid-feature-schema-id"
    ):
        client.unarchive_feature_schema_node(ontology.uid,
                                             'invalid-feature-schema-id')


def test_unarchive_feature_schema_node_for_non_existing_ontology(
        client, ontology):
    feature_schema_to_unarchive = ontology.normalized['tools'][0]
    with pytest.raises(Exception,
                       match="Failed to find ontology by id: invalid-ontology"):
        client.unarchive_feature_schema_node(
            'invalid-ontology', feature_schema_to_unarchive['featureSchemaId'])



import pytest


@pytest.mark.parametrize('data_rows', [3], indirect=True)
def test_dataset_export_v2(export_v2_test_helpers, dataset, data_rows):
    data_row_ids = [dr.uid for dr in data_rows]
    params = {"performance_details": False, "label_details": False}
    task_results = export_v2_test_helpers.run_dataset_export_v2_task(
        dataset, params=params)
    assert len(task_results) == len(data_row_ids)
    assert set([dr['data_row']['id'] for dr in task_results
               ]) == set(data_row_ids)

    # testing with a datarow ids filter
    datarow_filter_size = 2
    data_row_ids = [dr.uid for dr in data_rows]

    params = {"performance_details": False, "label_details": False}
    filters = {"data_row_ids": data_row_ids[:datarow_filter_size]}

    task_results = export_v2_test_helpers.run_dataset_export_v2_task(
        dataset, filters=filters, params=params)

    # only 2 datarows should be exported
    assert len(task_results) == datarow_filter_size
    # only filtered datarows should be exported
    assert set([dr['data_row']['id'] for dr in task_results
               ]) == set(data_row_ids[:datarow_filter_size])

    # testing with a global key and a datarow id filter
    datarow_filter_size = 2
    global_keys = [dr.global_key for dr in data_rows]

    params = {"performance_details": False, "label_details": False}
    filters = {"global_keys": global_keys[:datarow_filter_size]}

    task_results = export_v2_test_helpers.run_dataset_export_v2_task(
        dataset, filters=filters, params=params)

    # only 2 datarows should be exported
    assert len(task_results) == datarow_filter_size
    # only filtered datarows should be exported
    assert set([dr['data_row']['global_key'] for dr in task_results
               ]) == set(global_keys[:datarow_filter_size])



import pytest
from labelbox.utils import format_iso_datetime, format_iso_from_string


@pytest.mark.parametrize('datetime_str, expected_datetime_str',
                         [('2011-11-04T00:05:23Z', '2011-11-04T00:05:23Z'),
                          ('2011-11-04T00:05:23+00:00', '2011-11-04T00:05:23Z'),
                          ('2011-11-04T00:05:23+05:00', '2011-11-03T19:05:23Z'),
                          ('2011-11-04T00:05:23', '2011-11-04T00:05:23Z')])
def test_datetime_parsing(datetime_str, expected_datetime_str):
    # NOTE I would normally not take 'expected' using another function from sdk code, but in this case this is exactly the usage in _validate_parse_datetime
    assert format_iso_datetime(
        format_iso_from_string(datetime_str)) == expected_datetime_str



import pytest
from labelbox.exceptions import ResourceNotFoundError

def test_create_model_config(client, valid_model_id):
    model_config = client.create_model_config("model_config", valid_model_id, {"param": "value"})
    assert model_config.inference_params["param"] == "value"
    assert model_config.name == "model_config"
    assert model_config.model_id == valid_model_id


def test_delete_model_config(client, valid_model_id):
    model_config_id = client.create_model_config("model_config", valid_model_id, {"param": "value"})
    assert(client.delete_model_config(model_config_id.uid))

def test_delete_nonexistant_model_config(client):
    with pytest.raises(ResourceNotFoundError):
        client.delete_model_config("invalid_model_id")



from labelbox.data.annotation_types import Label, VideoObjectAnnotation
from labelbox.data.serialization.ndjson.converter import NDJsonConverter
from labelbox.data.annotation_types.geometry import Rectangle, Point
from labelbox.data.annotation_types import VideoData


def video_bbox_label():
    return Label(
        uid='cl1z52xwh00050fhcmfgczqvn',
        data=VideoData(
            uid="cklr9mr4m5iao0rb6cvxu4qbn",
            file_path=None,
            frames=None,
            url=
            "https://storage.labelbox.com/ckcz6bubudyfi0855o1dt1g9s%2F26403a22-604a-a38c-eeff-c2ed481fb40a-cat.mp4?Expires=1651677421050&KeyName=labelbox-assets-key-3&Signature=vF7gMyfHzgZdfbB8BHgd88Ws-Ms"
        ),
        annotations=[
            VideoObjectAnnotation(name='bbox toy',
                                  feature_schema_id='ckz38ofop0mci0z9i9w3aa9o4',
                                  extra={
                                      'value': 'bbox_toy',
                                      'instanceURI': None,
                                      'color': '#1CE6FF',
                                      'feature_id': 'cl1z52xw700000fhcayaqy0ev'
                                  },
                                  value=Rectangle(extra={},
                                                  start=Point(extra={},
                                                              x=70.0,
                                                              y=46.0),
                                                  end=Point(extra={},
                                                            x=454.0,
                                                            y=295.0)),
                                  classifications=[],
                                  frame=1,
                                  keyframe=True),
            VideoObjectAnnotation(name='bbox toy',
                                  feature_schema_id='ckz38ofop0mci0z9i9w3aa9o4',
                                  extra={
                                      'value': 'bbox_toy',
                                      'instanceURI': None,
                                      'color': '#1CE6FF',
                                      'feature_id': 'cl1z52xw700000fhcayaqy0ev'
                                  },
                                  value=Rectangle(extra={},
                                                  start=Point(extra={},
                                                              x=70.0,
                                                              y=42.5),
                                                  end=Point(extra={},
                                                            x=427.25,
                                                            y=308.25)),
                                  classifications=[],
                                  frame=2,
                                  keyframe=False),
            VideoObjectAnnotation(name='bbox toy',
                                  feature_schema_id='ckz38ofop0mci0z9i9w3aa9o4',
                                  extra={
                                      'value': 'bbox_toy',
                                      'instanceURI': None,
                                      'color': '#1CE6FF',
                                      'feature_id': 'cl1z52xw700000fhcayaqy0ev'
                                  },
                                  value=Rectangle(extra={},
                                                  start=Point(extra={},
                                                              x=70.0,
                                                              y=39.0),
                                                  end=Point(extra={},
                                                            x=400.5,
                                                            y=321.5)),
                                  classifications=[],
                                  frame=3,
                                  keyframe=False),
            VideoObjectAnnotation(name='bbox toy',
                                  feature_schema_id='ckz38ofop0mci0z9i9w3aa9o4',
                                  extra={
                                      'value': 'bbox_toy',
                                      'instanceURI': None,
                                      'color': '#1CE6FF',
                                      'feature_id': 'cl1z52xw700000fhcayaqy0ev'
                                  },
                                  value=Rectangle(extra={},
                                                  start=Point(extra={},
                                                              x=70.0,
                                                              y=35.5),
                                                  end=Point(extra={},
                                                            x=373.75,
                                                            y=334.75)),
                                  classifications=[],
                                  frame=4,
                                  keyframe=False),
            VideoObjectAnnotation(name='bbox toy',
                                  feature_schema_id='ckz38ofop0mci0z9i9w3aa9o4',
                                  extra={
                                      'value': 'bbox_toy',
                                      'instanceURI': None,
                                      'color': '#1CE6FF',
                                      'feature_id': 'cl1z52xw700000fhcayaqy0ev'
                                  },
                                  value=Rectangle(extra={},
                                                  start=Point(extra={},
                                                              x=70.0,
                                                              y=32.0),
                                                  end=Point(extra={},
                                                            x=347.0,
                                                            y=348.0)),
                                  classifications=[],
                                  frame=5,
                                  keyframe=True),
            VideoObjectAnnotation(name='bbox toy',
                                  feature_schema_id='ckz38ofop0mci0z9i9w3aa9o4',
                                  extra={
                                      'value': 'bbox_toy',
                                      'instanceURI': None,
                                      'color': '#1CE6FF',
                                      'feature_id': 'cl1z52xw700000fhcayaqy0ev'
                                  },
                                  value=Rectangle(extra={},
                                                  start=Point(extra={},
                                                              x=70.0,
                                                              y=132.0),
                                                  end=Point(extra={},
                                                            x=283.0,
                                                            y=348.0)),
                                  classifications=[],
                                  frame=9,
                                  keyframe=True),
            VideoObjectAnnotation(name='bbox toy',
                                  feature_schema_id='ckz38ofop0mci0z9i9w3aa9o4',
                                  extra={
                                      'value': 'bbox_toy',
                                      'instanceURI': None,
                                      'color': '#1CE6FF',
                                      'feature_id': 'cl1z52xw700000fhcayaqy0ev'
                                  },
                                  value=Rectangle(extra={},
                                                  start=Point(extra={},
                                                              x=70.0,
                                                              y=122.333),
                                                  end=Point(extra={},
                                                            x=295.5,
                                                            y=348.0)),
                                  classifications=[],
                                  frame=10,
                                  keyframe=False),
            VideoObjectAnnotation(name='bbox toy',
                                  feature_schema_id='ckz38ofop0mci0z9i9w3aa9o4',
                                  extra={
                                      'value': 'bbox_toy',
                                      'instanceURI': None,
                                      'color': '#1CE6FF',
                                      'feature_id': 'cl1z52xw700000fhcayaqy0ev'
                                  },
                                  value=Rectangle(extra={},
                                                  start=Point(extra={},
                                                              x=70.0,
                                                              y=112.667),
                                                  end=Point(extra={},
                                                            x=308.0,
                                                            y=348.0)),
                                  classifications=[],
                                  frame=11,
                                  keyframe=False),
            VideoObjectAnnotation(name='bbox toy',
                                  feature_schema_id='ckz38ofop0mci0z9i9w3aa9o4',
                                  extra={
                                      'value': 'bbox_toy',
                                      'instanceURI': None,
                                      'color': '#1CE6FF',
                                      'feature_id': 'cl1z52xw700000fhcayaqy0ev'
                                  },
                                  value=Rectangle(extra={},
                                                  start=Point(extra={},
                                                              x=70.0,
                                                              y=103.0),
                                                  end=Point(extra={},
                                                            x=320.5,
                                                            y=348.0)),
                                  classifications=[],
                                  frame=12,
                                  keyframe=False),
            VideoObjectAnnotation(name='bbox toy',
                                  feature_schema_id='ckz38ofop0mci0z9i9w3aa9o4',
                                  extra={
                                      'value': 'bbox_toy',
                                      'instanceURI': None,
                                      'color': '#1CE6FF',
                                      'feature_id': 'cl1z52xw700000fhcayaqy0ev'
                                  },
                                  value=Rectangle(extra={},
                                                  start=Point(extra={},
                                                              x=70.0,
                                                              y=93.333),
                                                  end=Point(extra={},
                                                            x=333.0,
                                                            y=348.0)),
                                  classifications=[],
                                  frame=13,
                                  keyframe=False),
            VideoObjectAnnotation(name='bbox toy',
                                  feature_schema_id='ckz38ofop0mci0z9i9w3aa9o4',
                                  extra={
                                      'value': 'bbox_toy',
                                      'instanceURI': None,
                                      'color': '#1CE6FF',
                                      'feature_id': 'cl1z52xw700000fhcayaqy0ev'
                                  },
                                  value=Rectangle(extra={},
                                                  start=Point(extra={},
                                                              x=70.0,
                                                              y=83.667),
                                                  end=Point(extra={},
                                                            x=345.5,
                                                            y=348.0)),
                                  classifications=[],
                                  frame=14,
                                  keyframe=False),
            VideoObjectAnnotation(name='bbox toy',
                                  feature_schema_id='ckz38ofop0mci0z9i9w3aa9o4',
                                  extra={
                                      'value': 'bbox_toy',
                                      'instanceURI': None,
                                      'color': '#1CE6FF',
                                      'feature_id': 'cl1z52xw700000fhcayaqy0ev'
                                  },
                                  value=Rectangle(extra={},
                                                  start=Point(extra={},
                                                              x=70.0,
                                                              y=74.0),
                                                  end=Point(extra={},
                                                            x=358.0,
                                                            y=348.0)),
                                  classifications=[],
                                  frame=15,
                                  keyframe=True),
            VideoObjectAnnotation(name='bbox toy',
                                  feature_schema_id='ckz38ofop0mci0z9i9w3aa9o4',
                                  extra={
                                      'value': 'bbox_toy',
                                      'instanceURI': None,
                                      'color': '#1CE6FF',
                                      'feature_id': 'cl1z52xw700000fhcayaqy0ev'
                                  },
                                  value=Rectangle(extra={},
                                                  start=Point(extra={},
                                                              x=70.0,
                                                              y=66.833),
                                                  end=Point(extra={},
                                                            x=387.333,
                                                            y=348.0)),
                                  classifications=[],
                                  frame=16,
                                  keyframe=False),
            VideoObjectAnnotation(name='bbox toy',
                                  feature_schema_id='ckz38ofop0mci0z9i9w3aa9o4',
                                  extra={
                                      'value': 'bbox_toy',
                                      'instanceURI': None,
                                      'color': '#1CE6FF',
                                      'feature_id': 'cl1z52xw700000fhcayaqy0ev'
                                  },
                                  value=Rectangle(extra={},
                                                  start=Point(extra={},
                                                              x=70.0,
                                                              y=59.667),
                                                  end=Point(extra={},
                                                            x=416.667,
                                                            y=348.0)),
                                  classifications=[],
                                  frame=17,
                                  keyframe=False),
            VideoObjectAnnotation(name='bbox toy',
                                  feature_schema_id='ckz38ofop0mci0z9i9w3aa9o4',
                                  extra={
                                      'value': 'bbox_toy',
                                      'instanceURI': None,
                                      'color': '#1CE6FF',
                                      'feature_id': 'cl1z52xw700000fhcayaqy0ev'
                                  },
                                  value=Rectangle(extra={},
                                                  start=Point(extra={},
                                                              x=70.0,
                                                              y=52.5),
                                                  end=Point(extra={},
                                                            x=446.0,
                                                            y=348.0)),
                                  classifications=[],
                                  frame=18,
                                  keyframe=False),
            VideoObjectAnnotation(name='bbox toy',
                                  feature_schema_id='ckz38ofop0mci0z9i9w3aa9o4',
                                  extra={
                                      'value': 'bbox_toy',
                                      'instanceURI': None,
                                      'color': '#1CE6FF',
                                      'feature_id': 'cl1z52xw700000fhcayaqy0ev'
                                  },
                                  value=Rectangle(extra={},
                                                  start=Point(extra={},
                                                              x=70.0,
                                                              y=45.333),
                                                  end=Point(extra={},
                                                            x=475.333,
                                                            y=348.0)),
                                  classifications=[],
                                  frame=19,
                                  keyframe=False),
            VideoObjectAnnotation(name='bbox toy',
                                  feature_schema_id='ckz38ofop0mci0z9i9w3aa9o4',
                                  extra={
                                      'value': 'bbox_toy',
                                      'instanceURI': None,
                                      'color': '#1CE6FF',
                                      'feature_id': 'cl1z52xw700000fhcayaqy0ev'
                                  },
                                  value=Rectangle(extra={},
                                                  start=Point(extra={},
                                                              x=70.0,
                                                              y=38.167),
                                                  end=Point(extra={},
                                                            x=504.667,
                                                            y=348.0)),
                                  classifications=[],
                                  frame=20,
                                  keyframe=False),
            VideoObjectAnnotation(name='bbox toy',
                                  feature_schema_id='ckz38ofop0mci0z9i9w3aa9o4',
                                  extra={
                                      'value': 'bbox_toy',
                                      'instanceURI': None,
                                      'color': '#1CE6FF',
                                      'feature_id': 'cl1z52xw700000fhcayaqy0ev'
                                  },
                                  value=Rectangle(extra={},
                                                  start=Point(extra={},
                                                              x=70.0,
                                                              y=31.0),
                                                  end=Point(extra={},
                                                            x=534.0,
                                                            y=348.0)),
                                  classifications=[],
                                  frame=21,
                                  keyframe=True),
            VideoObjectAnnotation(name='bbox toy',
                                  feature_schema_id='ckz38ofop0mci0z9i9w3aa9o4',
                                  extra={
                                      'value': 'bbox_toy',
                                      'instanceURI': None,
                                      'color': '#1CE6FF',
                                      'feature_id': 'cl1z52xw700000fhcayaqy0ev'
                                  },
                                  value=Rectangle(extra={},
                                                  start=Point(extra={},
                                                              x=70.0,
                                                              y=29.5),
                                                  end=Point(extra={},
                                                            x=543.0,
                                                            y=348.0)),
                                  classifications=[],
                                  frame=22,
                                  keyframe=False),
            VideoObjectAnnotation(name='bbox toy',
                                  feature_schema_id='ckz38ofop0mci0z9i9w3aa9o4',
                                  extra={
                                      'value': 'bbox_toy',
                                      'instanceURI': None,
                                      'color': '#1CE6FF',
                                      'feature_id': 'cl1z52xw700000fhcayaqy0ev'
                                  },
                                  value=Rectangle(extra={},
                                                  start=Point(extra={},
                                                              x=70.0,
                                                              y=28.0),
                                                  end=Point(extra={},
                                                            x=552.0,
                                                            y=348.0)),
                                  classifications=[],
                                  frame=23,
                                  keyframe=False),
            VideoObjectAnnotation(name='bbox toy',
                                  feature_schema_id='ckz38ofop0mci0z9i9w3aa9o4',
                                  extra={
                                      'value': 'bbox_toy',
                                      'instanceURI': None,
                                      'color': '#1CE6FF',
                                      'feature_id': 'cl1z52xw700000fhcayaqy0ev'
                                  },
                                  value=Rectangle(extra={},
                                                  start=Point(extra={},
                                                              x=70.0,
                                                              y=26.5),
                                                  end=Point(extra={},
                                                            x=561.0,
                                                            y=348.0)),
                                  classifications=[],
                                  frame=24,
                                  keyframe=False),
            VideoObjectAnnotation(name='bbox toy',
                                  feature_schema_id='ckz38ofop0mci0z9i9w3aa9o4',
                                  extra={
                                      'value': 'bbox_toy',
                                      'instanceURI': None,
                                      'color': '#1CE6FF',
                                      'feature_id': 'cl1z52xw700000fhcayaqy0ev'
                                  },
                                  value=Rectangle(extra={},
                                                  start=Point(extra={},
                                                              x=70.0,
                                                              y=25.0),
                                                  end=Point(extra={},
                                                            x=570.0,
                                                            y=348.0)),
                                  classifications=[],
                                  frame=25,
                                  keyframe=False),
            VideoObjectAnnotation(name='bbox toy',
                                  feature_schema_id='ckz38ofop0mci0z9i9w3aa9o4',
                                  extra={
                                      'value': 'bbox_toy',
                                      'instanceURI': None,
                                      'color': '#1CE6FF',
                                      'feature_id': 'cl1z52xw700000fhcayaqy0ev'
                                  },
                                  value=Rectangle(extra={},
                                                  start=Point(extra={},
                                                              x=70.0,
                                                              y=23.5),
                                                  end=Point(extra={},
                                                            x=579.0,
                                                            y=348.0)),
                                  classifications=[],
                                  frame=26,
                                  keyframe=False),
            VideoObjectAnnotation(name='bbox toy',
                                  feature_schema_id='ckz38ofop0mci0z9i9w3aa9o4',
                                  extra={
                                      'value': 'bbox_toy',
                                      'instanceURI': None,
                                      'color': '#1CE6FF',
                                      'feature_id': 'cl1z52xw700000fhcayaqy0ev'
                                  },
                                  value=Rectangle(extra={},
                                                  start=Point(extra={},
                                                              x=70.0,
                                                              y=22.0),
                                                  end=Point(extra={},
                                                            x=588.0,
                                                            y=348.0)),
                                  classifications=[],
                                  frame=27,
                                  keyframe=False),
            VideoObjectAnnotation(name='bbox toy',
                                  feature_schema_id='ckz38ofop0mci0z9i9w3aa9o4',
                                  extra={
                                      'value': 'bbox_toy',
                                      'instanceURI': None,
                                      'color': '#1CE6FF',
                                      'feature_id': 'cl1z52xw700000fhcayaqy0ev'
                                  },
                                  value=Rectangle(extra={},
                                                  start=Point(extra={},
                                                              x=70.0,
                                                              y=20.5),
                                                  end=Point(extra={},
                                                            x=597.0,
                                                            y=348.0)),
                                  classifications=[],
                                  frame=28,
                                  keyframe=False),
            VideoObjectAnnotation(name='bbox toy',
                                  feature_schema_id='ckz38ofop0mci0z9i9w3aa9o4',
                                  extra={
                                      'value': 'bbox_toy',
                                      'instanceURI': None,
                                      'color': '#1CE6FF',
                                      'feature_id': 'cl1z52xw700000fhcayaqy0ev'
                                  },
                                  value=Rectangle(extra={},
                                                  start=Point(extra={},
                                                              x=70.0,
                                                              y=19.0),
                                                  end=Point(extra={},
                                                            x=606.0,
                                                            y=348.0)),
                                  classifications=[],
                                  frame=29,
                                  keyframe=True)
        ],
        extra={
            'Created By':
                'jtso@labelbox.com',
            'Project Name':
                'Pictor Video',
            'Created At':
                '2022-04-14T15:11:19.000Z',
            'Updated At':
                '2022-04-14T15:11:21.064Z',
            'Seconds to Label':
                0.0,
            'Agreement':
                -1.0,
            'Benchmark Agreement':
                -1.0,
            'Benchmark ID':
                None,
            'Dataset Name':
                'cat',
            'Reviews': [],
            'View Label':
                'https://editor.labelbox.com?project=ckz38nsfd0lzq109bhq73est1&label=cl1z52xwh00050fhcmfgczqvn',
            'Has Open Issues':
                0.0,
            'Skipped':
                False,
            'media_type':
                'video',
            'Data Split':
                None
        })


def video_serialized_bbox_label():
    return {
        'uuid':
            'b24e672b-8f79-4d96-bf5e-b552ca0820d5',
        'dataRow': {
            'id': 'cklr9mr4m5iao0rb6cvxu4qbn'
        },
        'schemaId':
            'ckz38ofop0mci0z9i9w3aa9o4',
        'name':
            'bbox toy',
        'classifications': [],
        'segments': [{
            'keyframes': [{
                'frame': 1,
                'bbox': {
                    'top': 46.0,
                    'left': 70.0,
                    'height': 249.0,
                    'width': 384.0
                },
                'classifications': []
            }, {
                'frame': 5,
                'bbox': {
                    'top': 32.0,
                    'left': 70.0,
                    'height': 316.0,
                    'width': 277.0
                },
                'classifications': []
            }]
        }, {
            'keyframes': [{
                'frame': 9,
                'bbox': {
                    'top': 132.0,
                    'left': 70.0,
                    'height': 216.0,
                    'width': 213.0
                },
                'classifications': []
            }, {
                'frame': 15,
                'bbox': {
                    'top': 74.0,
                    'left': 70.0,
                    'height': 274.0,
                    'width': 288.0
                },
                'classifications': []
            }, {
                'frame': 21,
                'bbox': {
                    'top': 31.0,
                    'left': 70.0,
                    'height': 317.0,
                    'width': 464.0
                },
                'classifications': []
            }, {
                'frame': 29,
                'bbox': {
                    'top': 19.0,
                    'left': 70.0,
                    'height': 329.0,
                    'width': 536.0
                },
                'classifications': []
            }]
        }]
    }


def test_serialize_video_objects():
    label = video_bbox_label()
    serialized_labels = NDJsonConverter.serialize([label])
    label = next(serialized_labels)

    manual_label = video_serialized_bbox_label()

    for key in label.keys():
        # ignore uuid because we randomize if there was none
        if key != "uuid":
            assert label[key] == manual_label[key]

    assert len(label['segments']) == 2
    assert len(label['segments'][0]['keyframes']) == 2
    assert len(label['segments'][1]['keyframes']) == 4

    # #converts back only the keyframes. should be the sum of all prev segments
    deserialized_labels = NDJsonConverter.deserialize([label])
    label = next(deserialized_labels)
    assert len(label.annotations) == 6


def test_confidence_is_ignored():
    label = video_bbox_label()
    serialized_labels = NDJsonConverter.serialize([label])
    label = next(serialized_labels)
    label["confidence"] = 0.453
    label['segments'][0]["confidence"] = 0.453

    deserialized_labels = NDJsonConverter.deserialize([label])
    label = next(deserialized_labels)
    for annotation in label.annotations:
        assert annotation.confidence is None



from labelbox.schema.media_type import MediaType
from labelbox.schema.project import Project
import pytest

from labelbox import parser
from pytest_cases import parametrize, fixture_ref

from labelbox.exceptions import MALValidationError
from labelbox.schema.bulk_import_request import (NDChecklist, NDClassification,
                                                 NDMask, NDPolygon, NDPolyline,
                                                 NDRadio, NDRectangle, NDText,
                                                 NDTextEntity, NDTool,
                                                 _validate_ndjson)
"""
- These NDlabels are apart of bulkImportReqeust and should be removed once bulk import request is removed
"""

def test_classification_construction(checklist_inference, text_inference):
    checklist = NDClassification.build(checklist_inference[0])
    assert isinstance(checklist, NDChecklist)
    text = NDClassification.build(text_inference[0])
    assert isinstance(text, NDText)


@parametrize("inference, expected_type",
             [(fixture_ref('polygon_inference'), NDPolygon),
              (fixture_ref('rectangle_inference'), NDRectangle),
              (fixture_ref('line_inference'), NDPolyline),
              (fixture_ref('entity_inference'), NDTextEntity),
              (fixture_ref('segmentation_inference'), NDMask),
              (fixture_ref('segmentation_inference_rle'), NDMask),
              (fixture_ref('segmentation_inference_png'), NDMask)])
def test_tool_construction(inference, expected_type):
    assert isinstance(NDTool.build(inference[0]), expected_type)


def no_tool(text_inference, module_project):
    pred = text_inference[0].copy()
    #Missing key
    del pred['answer']
    with pytest.raises(MALValidationError):
        _validate_ndjson([pred], module_project)

@pytest.mark.parametrize(
    "configured_project",
    [MediaType.Text],
    indirect=True
)
def test_invalid_text(text_inference, configured_project):
    #and if it is not a string
    pred = text_inference[0].copy()
    #Extra and wrong key
    del pred['answer']
    pred['answers'] = []
    with pytest.raises(MALValidationError):
        _validate_ndjson([pred], configured_project)
    del pred['answers']

    #Invalid type
    pred['answer'] = []
    with pytest.raises(MALValidationError):
        _validate_ndjson([pred], configured_project)

    #Invalid type
    pred['answer'] = None
    with pytest.raises(MALValidationError):
        _validate_ndjson([pred], configured_project)


def test_invalid_checklist_item(checklist_inference,
                                module_project):
    #Only two points
    pred = checklist_inference[0].copy()
    pred['answers'] = [pred['answers'][0], pred['answers'][0]]
    #Duplicate schema ids
    with pytest.raises(MALValidationError):
        _validate_ndjson([pred], module_project)

    pred['answers'] = [{"name": "asdfg"}]
    with pytest.raises(MALValidationError):
        _validate_ndjson([pred], module_project)

    pred['answers'] = [{"schemaId": "1232132132"}]
    with pytest.raises(MALValidationError):
        _validate_ndjson([pred], module_project)

    pred['answers'] = [{}]
    with pytest.raises(MALValidationError):
        _validate_ndjson([pred], module_project)

    pred['answers'] = []
    with pytest.raises(MALValidationError):
        _validate_ndjson([pred], module_project)

    del pred['answers']
    with pytest.raises(MALValidationError):
        _validate_ndjson([pred], module_project)


def test_invalid_polygon(polygon_inference, module_project):
    #Only two points
    pred = polygon_inference[0].copy()
    pred['polygon'] = [{"x": 100, "y": 100}, {"x": 200, "y": 200}]
    with pytest.raises(MALValidationError):
        _validate_ndjson([pred], module_project)


@pytest.mark.parametrize(
    "configured_project",
    [MediaType.Text],
    indirect=True
)
def test_incorrect_entity(entity_inference, configured_project):
    entity = entity_inference[0].copy()
    #Location cannot be a list
    entity["location"] = [0, 10]
    with pytest.raises(MALValidationError):
        _validate_ndjson([entity], configured_project)

    entity["location"] = {"start": -1, "end": 5}
    with pytest.raises(MALValidationError):
        _validate_ndjson([entity], configured_project)

    entity["location"] = {"start": 15, "end": 5}
    with pytest.raises(MALValidationError):
        _validate_ndjson([entity], configured_project)


@pytest.mark.skip("Test wont work/fails randomly since projects have to have a media type and could be missing features from prediction list")
def test_all_validate_json(module_project, predictions):
    #Predictions contains one of each type of prediction.
    #These should be properly formatted and pass.
    _validate_ndjson(predictions[0], module_project)


def test_incorrect_line(line_inference, module_project):
    line = line_inference[0].copy()
    line["line"] = [line["line"][0]]  #Just one point
    with pytest.raises(MALValidationError):
        _validate_ndjson([line], module_project)


def test_incorrect_rectangle(rectangle_inference,
                             module_project):
    del rectangle_inference[0]['bbox']['top']
    with pytest.raises(MALValidationError):
        _validate_ndjson([rectangle_inference],
                         module_project)


def test_duplicate_tools(rectangle_inference, module_project):
    pred = rectangle_inference[0].copy()
    pred['polygon'] = [{"x": 100, "y": 100}, {"x": 200, "y": 200}]
    with pytest.raises(MALValidationError):
        _validate_ndjson([pred], module_project)


def test_invalid_feature_schema(module_project,
                                rectangle_inference):
    pred = rectangle_inference[0].copy()
    pred['schemaId'] = "blahblah"
    with pytest.raises(MALValidationError):
        _validate_ndjson([pred], module_project)


def test_name_only_feature_schema(module_project,
                                  rectangle_inference):
    pred = rectangle_inference[0].copy()
    _validate_ndjson([pred], module_project)


def test_schema_id_only_feature_schema(module_project,
                                       rectangle_inference):
    pred = rectangle_inference[0].copy()
    del pred['name']
    ontology = module_project.ontology().normalized["tools"]
    for tool in ontology:
        if tool["name"] == "bbox":
            feature_schema_id = tool["featureSchemaId"]
    pred["schemaId"] = feature_schema_id
    _validate_ndjson([pred], module_project)


def test_missing_feature_schema(module_project,
                                rectangle_inference):
    pred = rectangle_inference[0].copy()
    del pred['name']
    with pytest.raises(MALValidationError):
        _validate_ndjson([pred], module_project)


def test_validate_ndjson(tmp_path, configured_project):
    file_name = f"broken.ndjson"
    file_path = tmp_path / file_name
    with file_path.open("w") as f:
        f.write("test")

    with pytest.raises(ValueError):
        configured_project.upload_annotations(
            name="name", annotations=str(file_path), validate=True)


def test_validate_ndjson_uuid(tmp_path, configured_project,
                              predictions):
    file_name = f"repeat_uuid.ndjson"
    file_path = tmp_path / file_name
    repeat_uuid = predictions.copy()
    repeat_uuid[0]['uuid'] = 'test_uuid'
    repeat_uuid[1]['uuid'] = 'test_uuid'

    with file_path.open("w") as f:
        parser.dump(repeat_uuid, f)

    with pytest.raises(MALValidationError):
        configured_project.upload_annotations(
            name="name", validate=True, annotations=str(file_path))

    with pytest.raises(MALValidationError):
        configured_project.upload_annotations(
            name="name", validate=True, annotations=repeat_uuid)


@pytest.mark.parametrize("configured_project", [MediaType.Video], indirect=True)
def test_video_upload(video_checklist_inference,
                      configured_project):
    pred = video_checklist_inference[0].copy()
    _validate_ndjson([pred], configured_project)



import pytest

from labelbox.data.annotation_types import (Text, Point, Line,
                                            ClassificationAnnotation,
                                            ObjectAnnotation, TextEntity)
from labelbox.data.annotation_types.video import VideoObjectAnnotation
from labelbox.data.annotation_types.geometry.rectangle import Rectangle
from labelbox.data.annotation_types.video import VideoClassificationAnnotation
from labelbox.exceptions import ConfidenceNotSupportedException
from labelbox import pydantic_compat


def test_annotation():
    name = "line_feature"
    line = Line(points=[Point(x=1, y=2), Point(x=2, y=2)])
    classification = Text(answer="1234")

    annotation = ObjectAnnotation(
        value=line,
        name=name,
    )
    assert annotation.value.points[0].dict() == {'extra': {}, 'x': 1., 'y': 2.}
    assert annotation.name == name

    # Check ner
    ObjectAnnotation(
        value=TextEntity(start=10, end=12),
        name=name,
    )

    # Check classification
    ClassificationAnnotation(
        value=classification,
        name=name,
    )

    # Invalid subclass
    with pytest.raises(pydantic_compat.ValidationError):
        ObjectAnnotation(
            value=line,
            name=name,
            classifications=[line],
        )

    subclass = ClassificationAnnotation(value=classification, name=name)

    ObjectAnnotation(
        value=line,
        name=name,
        classifications=[subclass],
    )


def test_video_annotations():
    name = "line_feature"
    line = Line(points=[Point(x=1, y=2), Point(x=2, y=2)])

    # Wrong type
    with pytest.raises(pydantic_compat.ValidationError):
        VideoClassificationAnnotation(value=line, name=name, frame=1)

    # Missing frames
    with pytest.raises(pydantic_compat.ValidationError):
        VideoClassificationAnnotation(value=line, name=name)

    VideoObjectAnnotation(value=line, name=name, keyframe=True, frame=2)


def test_confidence_for_video_is_not_supported():
    with pytest.raises(ConfidenceNotSupportedException):
        VideoObjectAnnotation(name='bbox toy',
                              feature_schema_id='ckz38ofop0mci0z9i9w3aa9o4',
                              extra={
                                  'value': 'bbox_toy',
                                  'instanceURI': None,
                                  'color': '#1CE6FF',
                                  'feature_id': 'cl1z52xw700000fhcayaqy0ev'
                              },
                              value=Rectangle(extra={},
                                              start=Point(extra={},
                                                          x=70.0,
                                                          y=26.5),
                                              end=Point(extra={},
                                                        x=561.0,
                                                        y=348.0)),
                              classifications=[],
                              frame=24,
                              keyframe=False,
                              confidence=0.3434),


def test_confidence_value_range_validation():
    name = "line_feature"
    line = Line(points=[Point(x=1, y=2), Point(x=2, y=2)])

    with pytest.raises(ValueError) as e:
        ObjectAnnotation(value=line, name=name, confidence=14)
    assert e.value.errors()[0]['msg'] == 'must be a number within [0,1] range'



from datetime import datetime, timezone, timedelta
import pytest
import os


@pytest.mark.skipif(
    condition=os.environ['LABELBOX_TEST_ENVIRON'] == "onprem",
    reason="longer runtime than expected for onprem. unskip when resolved.")
def test_labeler_performance(configured_project_with_label):
    project, _, _, _ = configured_project_with_label

    labeler_performance = list(project.labeler_performance())
    assert len(labeler_performance) == 1
    my_performance = labeler_performance[0]
    assert my_performance.user == project.client.get_user()
    assert my_performance.count == 1
    assert isinstance(my_performance.last_activity_time, datetime)
    now_utc = datetime.now().astimezone(timezone.utc)
    assert timedelta(0) < now_utc - my_performance.last_activity_time < \
        timedelta(seconds=60)



import json
import random
import threading
from tempfile import NamedTemporaryFile
from typing import List, Dict, Any

import pytest

import labelbox.exceptions
from labelbox import Client, Dataset, DataRow
from labelbox.schema.embedding import Embedding


def test_get_embedding_by_id(client: Client, embedding: Embedding):
    e = client.get_embedding_by_id(embedding.id)
    assert e.id == embedding.id


def test_get_embedding_by_name(client: Client, embedding: Embedding):
    e = client.get_embedding_by_name(embedding.name)
    assert e.name == embedding.name


def test_get_embedding_by_name_not_found(client: Client):
    with pytest.raises(labelbox.exceptions.ResourceNotFoundError):
        client.get_embedding_by_name("does-not-exist")


def test_get_embeddings(client: Client, embedding: Embedding):
    embeddings = client.get_embeddings()
    assert len(embeddings) > 0


@pytest.mark.parametrize('data_rows', [10], indirect=True)
def test_import_vectors_from_file(data_rows: List[DataRow],
                                  embedding: Embedding):
    vector = [random.uniform(1.0, 2.0) for _ in range(embedding.dims)]
    event = threading.Event()

    def callback(_: Dict[str, Any]):
        event.set()

    with NamedTemporaryFile(mode="w+") as fp:
        lines = [
            json.dumps({
                "id": dr.uid,
                "vector": vector
            }) for dr in data_rows
        ]
        fp.writelines(lines)
        fp.flush()
        embedding.import_vectors_from_file(fp.name, callback)

    assert event.wait(10.0)  # seconds


def test_get_imported_vector_count(dataset: Dataset, embedding: Embedding):
    assert embedding.get_imported_vector_count() == 0

    vector = [random.uniform(1.0, 2.0) for _ in range(embedding.dims)]
    dataset.create_data_row(row_data="foo",
                            embeddings=[{
                                "embedding_id": embedding.id,
                                "vector": vector
                            }])

    assert embedding.get_imported_vector_count() == 1



import sys
from typing import Generic, TypeVar, Any

from typing_extensions import Annotated
from packaging import version
import numpy as np

from labelbox import pydantic_compat

Cuid = Annotated[str, pydantic_compat.Field(min_length=25, max_length=25)]

DType = TypeVar('DType')
DShape = TypeVar('DShape')


class _TypedArray(np.ndarray, Generic[DType, DShape]):

    @classmethod
    def __get_validators__(cls):
        yield cls.validate

    @classmethod
    def validate(cls, val, field: pydantic_compat.ModelField):
        if not isinstance(val, np.ndarray):
            raise TypeError(f"Expected numpy array. Found {type(val)}")

        if sys.version_info.minor > 6:
            actual_dtype = field.sub_fields[-1].type_.__args__[0]
        else:
            actual_dtype = field.sub_fields[-1].type_.__values__[0]

        if val.dtype != actual_dtype:
            raise TypeError(
                f"Expected numpy array have type {actual_dtype}. Found {val.dtype}"
            )
        return val


if version.parse(np.__version__) >= version.parse('1.25.0'):
    from typing import GenericAlias
    TypedArray = GenericAlias(_TypedArray, (Any, DType))
elif version.parse(np.__version__) >= version.parse('1.23.0'):
    from numpy._typing import _GenericAlias
    TypedArray = _GenericAlias(_TypedArray, (Any, DType))
elif version.parse('1.22.0') <= version.parse(
        np.__version__) < version.parse('1.23.0'):
    from numpy.typing import _GenericAlias
    TypedArray = _GenericAlias(_TypedArray, (Any, DType))
else:
    TypedArray = _TypedArray[Any, DType]



import logging
import os
import urllib.request
from typing import Callable, Dict, Generator, Optional, Tuple
from typing_extensions import Literal
from uuid import uuid4

import cv2
import numpy as np
from google.api_core import retry

from .base_data import BaseData
from ..types import TypedArray

from labelbox import pydantic_compat

logger = logging.getLogger(__name__)


class VideoData(BaseData):
    """
    Represents video
    """
    file_path: Optional[str] = None
    url: Optional[str] = None
    frames: Optional[Dict[int, TypedArray[Literal['uint8']]]] = None

    def load_frames(self, overwrite: bool = False) -> None:
        """
        Loads all frames into memory at once in order to access in non-sequential order.
        This will use a lot of memory, especially for longer videos

        Args:
            overwrite: Replace existing frames
        """
        if self.frames and not overwrite:
            return

        for count, frame in self.frame_generator():
            if self.frames is None:
                self.frames = {}
            self.frames[count] = frame

    @property
    def value(self):
        return self.frame_generator()

    def frame_generator(
            self,
            cache_frames=False,
            download_dir='/tmp'
    ) -> Generator[Tuple[int, np.ndarray], None, None]:
        """
        A generator for accessing individual frames in a video.

        Args:
            cache_frames (bool): Whether or not to cache frames while iterating through the video.
            download_dir (str): Directory to save the video to. Defaults to `/tmp` dir
        """
        if self.frames is not None:
            for idx, frame in self.frames.items():
                yield idx, frame
            return
        elif self.url and not self.file_path:
            file_path = os.path.join(download_dir, f"{uuid4()}.mp4")
            logger.info("Downloading the video locally to %s", file_path)
            self.fetch_remote(file_path)
            self.file_path = file_path

        vidcap = cv2.VideoCapture(self.file_path)

        success, frame = vidcap.read()
        count = 0
        if cache_frames:
            self.frames = {}
        while success:
            frame = frame[:, :, ::-1]
            yield count, frame
            if cache_frames:
                self.frames[count] = frame
            success, frame = vidcap.read()
            count += 1

    def __getitem__(self, idx: int) -> np.ndarray:
        if self.frames is None:
            raise ValueError(
                "Cannot select by index without iterating over the entire video or loading all frames."
            )
        return self.frames[idx]

    def set_fetch_fn(self, fn):
        object.__setattr__(self, 'fetch_remote', lambda: fn(self))

    @retry.Retry(deadline=15.)
    def fetch_remote(self, local_path) -> None:
        """
        Method for downloading data from self.url

        If url is not publicly accessible or requires another access pattern
        simply override this function

        Args:
            local_path: Where to save the thing too.
        """
        urllib.request.urlretrieve(self.url, local_path)

    @retry.Retry(deadline=15.)
    def create_url(self, signer: Callable[[bytes], str]) -> None:
        """
        Utility for creating a url from any of the other video references.

        Args:
            signer: A function that accepts bytes and returns a signed url.
        Returns:
            url for the video
        """
        if self.url is not None:
            return self.url
        elif self.file_path is not None:
            with open(self.file_path, 'rb') as file:
                self.url = signer(file.read())
        elif self.frames is not None:
            self.file_path = self.frames_to_video(self.frames)
            self.url = self.create_url(signer)
        else:
            raise ValueError("One of url, file_path, frames must not be None.")
        return self.url

    def frames_to_video(self,
                        frames: Dict[int, np.ndarray],
                        fps=20,
                        save_dir='/tmp') -> str:
        """
        Compresses the data by converting a set of individual frames to a single video.

        """
        file_path = os.path.join(save_dir, f"{uuid4()}.mp4")
        out = None
        for key in frames.keys():
            frame = frames[key]
            if out is None:
                out = cv2.VideoWriter(file_path,
                                      cv2.VideoWriter_fourcc(*'MP4V'), fps,
                                      frame.shape[:2])
            out.write(frame)
        if out is None:
            return
        out.release()
        return file_path

    @pydantic_compat.root_validator
    def validate_data(cls, values):
        file_path = values.get("file_path")
        url = values.get("url")
        frames = values.get("frames")
        uid = values.get("uid")
        global_key = values.get("global_key")

        if uid == file_path == frames == url == global_key == None:
            raise ValueError(
                "One of `file_path`, `frames`, `uid`, `global_key` or `url` required."
            )
        return values

    def __repr__(self) -> str:
        return  f"VideoData(file_path={self.file_path}," \
                f"frames={'...' if self.frames is not None else None}," \
                f"url={self.url})"

    class Config:
        # Required for discriminating between data types
        extra = 'forbid'



from datetime import datetime

import pytest
import uuid

from labelbox import Dataset
from labelbox.exceptions import MalformedQueryException
from labelbox.schema.identifiables import GlobalKeys, UniqueIds
from labelbox.schema.data_row_metadata import DataRowMetadataField, DataRowMetadata, DataRowMetadataKind, DataRowMetadataOntology, _parse_metadata_schema

INVALID_SCHEMA_ID = "1" * 25
FAKE_SCHEMA_ID = "0" * 25
FAKE_DATAROW_ID = "D" * 25
SPLIT_SCHEMA_ID = "cko8sbczn0002h2dkdaxb5kal"
TRAIN_SPLIT_ID = "cko8sbscr0003h2dk04w86hof"
TEST_SPLIT_ID = "cko8scbz70005h2dkastwhgqt"
TEXT_SCHEMA_ID = "cko8s9r5v0001h2dk9elqdidh"
CAPTURE_DT_SCHEMA_ID = "cko8sdzv70006h2dk8jg64zvb"
CUSTOM_TEXT_SCHEMA_NAME = 'custom_text'

FAKE_NUMBER_FIELD = {
    "id": FAKE_SCHEMA_ID,
    "name": "number",
    "kind": 'CustomMetadataNumber',
    "reserved": False
}


@pytest.fixture
def mdo(client):
    mdo = client.get_data_row_metadata_ontology()
    try:
        mdo.create_schema(CUSTOM_TEXT_SCHEMA_NAME, DataRowMetadataKind.string)
    except MalformedQueryException:
        # Do nothing if already exists
        pass
    mdo._raw_ontology = mdo._get_ontology()
    mdo._raw_ontology.append(FAKE_NUMBER_FIELD)
    mdo._build_ontology()
    yield mdo


@pytest.fixture
def big_dataset(dataset: Dataset, image_url):
    task = dataset.create_data_rows([
        {
            "row_data": image_url,
            "external_id": "my-image"
        },
    ] * 5)
    task.wait_till_done()

    yield dataset


def make_metadata(dr_id: str = None, gk: str = None) -> DataRowMetadata:
    msg = "A message"
    time = datetime.utcnow()

    metadata = DataRowMetadata(
        global_key=gk,
        data_row_id=dr_id,
        fields=[
            DataRowMetadataField(schema_id=SPLIT_SCHEMA_ID,
                                 value=TEST_SPLIT_ID),
            DataRowMetadataField(schema_id=CAPTURE_DT_SCHEMA_ID, value=time),
            DataRowMetadataField(schema_id=TEXT_SCHEMA_ID, value=msg),
        ])
    return metadata


def make_named_metadata(dr_id) -> DataRowMetadata:
    msg = "A message"
    time = datetime.utcnow()

    metadata = DataRowMetadata(data_row_id=dr_id,
                               fields=[
                                   DataRowMetadataField(name='split',
                                                        value=TEST_SPLIT_ID),
                                   DataRowMetadataField(name='captureDateTime',
                                                        value=time),
                                   DataRowMetadataField(
                                       name=CUSTOM_TEXT_SCHEMA_NAME, value=msg),
                               ])
    return metadata


@pytest.mark.skip(reason="broken export v1 api, to be retired soon")
def test_export_empty_metadata(client, configured_project_with_label,
                               wait_for_data_row_processing):
    project, _, data_row, _ = configured_project_with_label
    data_row = wait_for_data_row_processing(client, data_row)
    
    export_task = project.export(params={"metadata_fields": True})
    export_task.wait_till_done()
    stream = export_task.get_buffered_stream()
    data_row = [data_row.json for data_row in stream][0]
    
    assert data_row["metadata_fields"] == []


def test_bulk_export_datarow_metadata(data_row, mdo: DataRowMetadataOntology):
    metadata = make_metadata(data_row.uid)
    mdo.bulk_upsert([metadata])
    exported = mdo.bulk_export([data_row.uid])
    assert exported[0].global_key == data_row.global_key
    assert exported[0].data_row_id == data_row.uid
    assert len([field for field in exported[0].fields]) == 3

    exported = mdo.bulk_export(UniqueIds([data_row.uid]))
    assert exported[0].global_key == data_row.global_key
    assert exported[0].data_row_id == data_row.uid
    assert len([field for field in exported[0].fields]) == 3

    exported = mdo.bulk_export(GlobalKeys([data_row.global_key]))
    assert exported[0].global_key == data_row.global_key
    assert exported[0].data_row_id == data_row.uid
    assert len([field for field in exported[0].fields]) == 3


def test_get_datarow_metadata_ontology(mdo):
    assert len(mdo.fields)
    assert len(mdo.reserved_fields)
    # two are created by mdo fixture but there may be more
    assert len(mdo.custom_fields) >= 2

    split = mdo.reserved_by_name["split"]["train"]

    assert DataRowMetadata(
        data_row_id=FAKE_DATAROW_ID,
        fields=[
            DataRowMetadataField(
                schema_id=mdo.reserved_by_name["captureDateTime"].uid,
                value=datetime.utcnow(),
            ),
            DataRowMetadataField(schema_id=split.parent, value=split.uid),
            DataRowMetadataField(schema_id=mdo.reserved_by_name["tag"].uid,
                                 value="hello-world"),
        ])


def test_bulk_upsert_datarow_metadata(data_row, mdo: DataRowMetadataOntology):
    metadata = make_metadata(data_row.uid)
    mdo.bulk_upsert([metadata])
    exported = mdo.bulk_export([data_row.uid])
    assert len(exported)
    assert len([field for field in exported[0].fields]) == 3


def test_bulk_upsert_datarow_metadata_by_globalkey(
        data_rows, mdo: DataRowMetadataOntology):
    global_keys = [data_row.global_key for data_row in data_rows]
    metadata = [make_metadata(gk=global_key) for global_key in global_keys]
    errors = mdo.bulk_upsert(metadata)
    assert len(errors) == 0


@pytest.mark.slow
def test_large_bulk_upsert_datarow_metadata(big_dataset, mdo):
    metadata = []
    data_row_ids = [dr.uid for dr in big_dataset.data_rows()]
    for data_row_id in data_row_ids:
        metadata.append(make_metadata(data_row_id))
    errors = mdo.bulk_upsert(metadata)
    assert len(errors) == 0

    metadata_lookup = {
        metadata.data_row_id: metadata
        for metadata in mdo.bulk_export(data_row_ids)
    }
    for data_row_id in data_row_ids:
        assert len([f for f in metadata_lookup.get(data_row_id).fields
                   ]), metadata_lookup.get(data_row_id).fields


def test_upsert_datarow_metadata_by_name(data_row, mdo):
    metadata = [make_named_metadata(data_row.uid)]
    errors = mdo.bulk_upsert(metadata)
    assert len(errors) == 0

    metadata_lookup = {
        metadata.data_row_id: metadata
        for metadata in mdo.bulk_export([data_row.uid])
    }
    assert len([f for f in metadata_lookup.get(data_row.uid).fields
               ]), metadata_lookup.get(data_row.uid).fields


def test_upsert_datarow_metadata_option_by_name(data_row, mdo):
    metadata = DataRowMetadata(data_row_id=data_row.uid,
                               fields=[
                                   DataRowMetadataField(name='split',
                                                        value='test'),
                               ])
    errors = mdo.bulk_upsert([metadata])
    assert len(errors) == 0

    datarows = mdo.bulk_export([data_row.uid])
    assert len(datarows[0].fields) == 1
    metadata = datarows[0].fields[0]
    assert metadata.schema_id == SPLIT_SCHEMA_ID
    assert metadata.name == 'test'
    assert metadata.value == TEST_SPLIT_ID


def test_upsert_datarow_metadata_option_by_incorrect_name(data_row, mdo):
    metadata = DataRowMetadata(data_row_id=data_row.uid,
                               fields=[
                                   DataRowMetadataField(name='split',
                                                        value='test1'),
                               ])
    with pytest.raises(KeyError):
        mdo.bulk_upsert([metadata])


def test_raise_enum_upsert_schema_error(data_row, mdo):
    """Setting an option id as the schema id will raise a Value Error"""

    metadata = DataRowMetadata(data_row_id=data_row.uid,
                               fields=[
                                   DataRowMetadataField(schema_id=TEST_SPLIT_ID,
                                                        value=SPLIT_SCHEMA_ID),
                               ])
    with pytest.raises(ValueError):
        mdo.bulk_upsert([metadata])


def test_upsert_non_existent_schema_id(data_row, mdo):
    """Raise error on non-existent schema id"""
    metadata = DataRowMetadata(data_row_id=data_row.uid,
                               fields=[
                                   DataRowMetadataField(
                                       schema_id=INVALID_SCHEMA_ID,
                                       value="message"),
                               ])
    with pytest.raises(ValueError):
        mdo.bulk_upsert([metadata])


def test_parse_raw_metadata(mdo):
    example = {
        'dataRowId':
            'ckr6kkfx801ui0yrtg9fje8xh',
        'globalKey':
            'global-key-1',
        'fields': [
            {
                'schemaId': 'cko8s9r5v0001h2dk9elqdidh',
                'value': 'my-new-message'
            },
            {
                'schemaId': 'cko8sbczn0002h2dkdaxb5kal',
                'value': {}
            },
            {
                'schemaId': 'cko8sbscr0003h2dk04w86hof',
                'value': {}
            },
            {
                'schemaId': 'cko8sdzv70006h2dk8jg64zvb',
                'value': '2021-07-20T21:41:14.606710Z'
            },
            {
                'schemaId': FAKE_SCHEMA_ID,
                'value': 0.5
            },
        ]
    }

    parsed = mdo.parse_metadata([example])
    assert len(parsed) == 1
    for row in parsed:
        assert row.data_row_id == example["dataRowId"]
        assert row.global_key == example["globalKey"]
        assert len(row.fields) == 4

    for row in parsed:
        for field in row.fields:
            assert mdo._parse_upsert(field)


def test_parse_raw_metadata_fields(mdo):
    example = [
        {
            'schemaId': 'cko8s9r5v0001h2dk9elqdidh',
            'value': 'my-new-message'
        },
        {
            'schemaId': 'cko8sbczn0002h2dkdaxb5kal',
            'value': {}
        },
        {
            'schemaId': 'cko8sbscr0003h2dk04w86hof',
            'value': {}
        },
        {
            'schemaId': 'cko8sdzv70006h2dk8jg64zvb',
            'value': '2021-07-20T21:41:14.606710Z'
        },
        {
            'schemaId': FAKE_SCHEMA_ID,
            'value': 0.5
        },
    ]

    parsed = mdo.parse_metadata_fields(example)
    assert len(parsed) == 4

    for field in parsed:
        assert mdo._parse_upsert(field)


def test_parse_metadata_schema():
    unparsed = {
        'id':
            'cl467a4ec0046076g7s9yheoa',
        'name':
            'enum metadata',
        'kind':
            'CustomMetadataEnum',
        'options': [{
            'id': 'cl467a4ec0047076ggjneeruy',
            'name': 'option1',
            'kind': 'CustomMetadataEnumOption'
        }, {
            'id': 'cl4qa31u0009e078p5m280jer',
            'name': 'option2',
            'kind': 'CustomMetadataEnumOption'
        }]
    }
    parsed = _parse_metadata_schema(unparsed)
    assert parsed.uid == 'cl467a4ec0046076g7s9yheoa'
    assert parsed.name == 'enum metadata'
    assert parsed.kind == DataRowMetadataKind.enum
    assert len(parsed.options) == 2
    assert parsed.options[0].uid == 'cl467a4ec0047076ggjneeruy'
    assert parsed.options[0].kind == DataRowMetadataKind.option


def test_create_schema(mdo):
    metadata_name = str(uuid.uuid4())
    created_schema = mdo.create_schema(metadata_name, DataRowMetadataKind.enum,
                                       ["option 1", "option 2"])
    assert created_schema.name == metadata_name
    assert created_schema.kind == DataRowMetadataKind.enum
    assert len(created_schema.options) == 2
    assert created_schema.options[0].name == "option 1"
    mdo.delete_schema(metadata_name)


def test_update_schema(mdo):
    metadata_name = str(uuid.uuid4())
    created_schema = mdo.create_schema(metadata_name, DataRowMetadataKind.enum,
                                       ["option 1", "option 2"])
    updated_schema = mdo.update_schema(metadata_name,
                                       f"{metadata_name}_updated")
    assert updated_schema.name == f"{metadata_name}_updated"
    assert updated_schema.uid == created_schema.uid
    assert updated_schema.kind == DataRowMetadataKind.enum
    mdo.delete_schema(f"{metadata_name}_updated")


def test_update_enum_options(mdo):
    metadata_name = str(uuid.uuid4())
    created_schema = mdo.create_schema(metadata_name, DataRowMetadataKind.enum,
                                       ["option 1", "option 2"])
    updated_schema = mdo.update_enum_option(metadata_name, "option 1",
                                            "option 3")
    assert updated_schema.name == metadata_name
    assert updated_schema.uid == created_schema.uid
    assert updated_schema.kind == DataRowMetadataKind.enum
    assert updated_schema.options[0].uid == created_schema.options[0].uid
    assert updated_schema.options[0].name == "option 3"
    mdo.delete_schema(metadata_name)


def test_delete_schema(mdo):
    metadata_name = str(uuid.uuid4())
    created_schema = mdo.create_schema(metadata_name,
                                       DataRowMetadataKind.string)
    status = mdo.delete_schema(created_schema.name)
    mdo.refresh_ontology()
    assert status
    assert metadata_name not in mdo.custom_by_name


@pytest.mark.parametrize('datetime_str',
                         ['2011-11-04T00:05:23Z', '2011-11-04T00:05:23+00:00'])
def test_upsert_datarow_date_metadata(data_row, mdo, datetime_str):
    metadata = [
        DataRowMetadata(data_row_id=data_row.uid,
                        fields=[
                            DataRowMetadataField(name='captureDateTime',
                                                 value=datetime_str),
                        ])
    ]
    errors = mdo.bulk_upsert(metadata)
    assert len(errors) == 0

    metadata = mdo.bulk_export([data_row.uid])
    assert f"{metadata[0].fields[0].value}" == "2011-11-04 00:05:23+00:00"


@pytest.mark.parametrize('datetime_str',
                         ['2011-11-04T00:05:23Z', '2011-11-04T00:05:23+00:00'])
def test_create_data_row_with_metadata(dataset, image_url, datetime_str):
    client = dataset.client
    assert len(list(dataset.data_rows())) == 0

    metadata_fields = [
        DataRowMetadataField(name='captureDateTime', value=datetime_str)
    ]

    data_row = dataset.create_data_row(row_data=image_url,
                                       metadata_fields=metadata_fields)

    retrieved_data_row = client.get_data_row(data_row.uid)
    assert f"{retrieved_data_row.metadata[0].value}" == "2011-11-04 00:05:23+00:00"



from labelbox import pydantic_compat
from pathlib import Path


class PathSerializerMixin(pydantic_compat.BaseModel):

    def dict(self, *args, **kwargs):
        res = super().dict(*args, **kwargs)
        return {k: str(v) if isinstance(v, Path) else v for k, v in res.items()}



# type: ignore
from datetime import datetime
from copy import deepcopy
from enum import Enum
from itertools import chain
import warnings

from typing import List, Optional, Dict, Union, Callable, Type, Any, Generator, overload

from labelbox import pydantic_compat
from labelbox.schema.identifiables import DataRowIdentifiers, UniqueIds
from labelbox.schema.identifiable import UniqueId, GlobalKey

from labelbox.schema.ontology import SchemaId
from labelbox.utils import _CamelCaseMixin, camel_case, format_iso_datetime, format_iso_from_string


class DataRowMetadataKind(Enum):
    number = "CustomMetadataNumber"
    datetime = "CustomMetadataDateTime"
    enum = "CustomMetadataEnum"
    string = "CustomMetadataString"
    option = "CustomMetadataEnumOption"
    embedding = "CustomMetadataEmbedding"


# Metadata schema
class DataRowMetadataSchema(pydantic_compat.BaseModel):
    uid: SchemaId
    name: pydantic_compat.constr(strip_whitespace=True,
                                 min_length=1,
                                 max_length=100)
    reserved: bool
    kind: DataRowMetadataKind
    options: Optional[List["DataRowMetadataSchema"]]
    parent: Optional[SchemaId]


DataRowMetadataSchema.update_forward_refs()

Embedding: Type[List[float]] = pydantic_compat.conlist(float,
                                                       min_items=128,
                                                       max_items=128)
String: Type[str] = pydantic_compat.constr(max_length=4096)


# Metadata base class
class DataRowMetadataField(_CamelCaseMixin):
    # One of `schema_id` or `name` must be provided. If `schema_id` is not provided, it is
    # inferred from `name`
    schema_id: Optional[SchemaId] = None
    name: Optional[str] = None
    # value is of type `Any` so that we do not improperly coerce the value to the wrong tpye
    # Additional validation is performed before upload using the schema information
    value: Any


class DataRowMetadata(_CamelCaseMixin):
    global_key: Optional[str] = None
    data_row_id: Optional[str] = None
    fields: List[DataRowMetadataField]


class DeleteDataRowMetadata(_CamelCaseMixin):
    data_row_id: Union[str, UniqueId, GlobalKey]
    fields: List[SchemaId]

    class Config:
        arbitrary_types_allowed = True


class DataRowMetadataBatchResponse(_CamelCaseMixin):
    global_key: Optional[str] = None
    data_row_id: Optional[str] = None
    error: Optional[str] = None
    fields: List[Union[DataRowMetadataField, SchemaId]]


# --- Batch GraphQL Objects ---
# Don't want to crowd the name space with internals


# Bulk upsert values
class _UpsertDataRowMetadataInput(_CamelCaseMixin):
    schema_id: str
    value: Any


# Batch of upsert values for a datarow
class _UpsertBatchDataRowMetadata(_CamelCaseMixin):
    global_key: Optional[str] = None
    data_row_id: Optional[str] = None
    fields: List[_UpsertDataRowMetadataInput]


class _DeleteBatchDataRowMetadata(_CamelCaseMixin):
    data_row_identifier: Union[UniqueId, GlobalKey]
    schema_ids: List[SchemaId]

    class Config:
        arbitrary_types_allowed = True
        alias_generator = camel_case

    def dict(self, *args, **kwargs):
        res = super().dict(*args, **kwargs)
        if 'data_row_identifier' in res.keys():
            key = 'data_row_identifier'
            id_type_key = 'id_type'
        else:
            key = 'dataRowIdentifier'
            id_type_key = 'idType'
        data_row_identifier = res.pop(key)
        res[key] = {
            "id": data_row_identifier.key,
            id_type_key: data_row_identifier.id_type
        }
        return res


_BatchInputs = Union[List[_UpsertBatchDataRowMetadata],
                     List[_DeleteBatchDataRowMetadata]]
_BatchFunction = Callable[[_BatchInputs], List[DataRowMetadataBatchResponse]]


class _UpsertCustomMetadataSchemaEnumOptionInput(_CamelCaseMixin):
    id: Optional[SchemaId]
    name: pydantic_compat.constr(strip_whitespace=True,
                                 min_length=1,
                                 max_length=100)
    kind: str


class _UpsertCustomMetadataSchemaInput(_CamelCaseMixin):
    id: Optional[SchemaId]
    name: pydantic_compat.constr(strip_whitespace=True,
                                 min_length=1,
                                 max_length=100)
    kind: str
    options: Optional[List[_UpsertCustomMetadataSchemaEnumOptionInput]]


class DataRowMetadataOntology:
    """ Ontology for data row metadata

    Metadata provides additional context for a data rows. Metadata is broken into two classes
    reserved and custom. Reserved fields are defined by Labelbox and used for creating
    specific experiences in the platform.

    >>> mdo = client.get_data_row_metadata_ontology()

    """

    def __init__(self, client):

        self._client = client
        self._batch_size = 50  # used for uploads and deletes

        self._raw_ontology = self._get_ontology()
        self._build_ontology()

    def _build_ontology(self):
        # all fields
        self.fields = self._parse_ontology(self._raw_ontology)
        self.fields_by_id = self._make_id_index(self.fields)

        # reserved fields
        self.reserved_fields: List[DataRowMetadataSchema] = [
            f for f in self.fields if f.reserved
        ]
        self.reserved_by_id = self._make_id_index(self.reserved_fields)
        self.reserved_by_name: Dict[str, Union[DataRowMetadataSchema, Dict[
            str, DataRowMetadataSchema]]] = self._make_name_index(
                self.reserved_fields)
        self.reserved_by_name_normalized: Dict[
            str, DataRowMetadataSchema] = self._make_normalized_name_index(
                self.reserved_fields)

        # custom fields
        self.custom_fields: List[DataRowMetadataSchema] = [
            f for f in self.fields if not f.reserved
        ]
        self.custom_by_id = self._make_id_index(self.custom_fields)
        self.custom_by_name: Dict[str, Union[DataRowMetadataSchema, Dict[
            str,
            DataRowMetadataSchema]]] = self._make_name_index(self.custom_fields)
        self.custom_by_name_normalized: Dict[
            str, DataRowMetadataSchema] = self._make_normalized_name_index(
                self.custom_fields)

    @staticmethod
    def _lookup_in_index_by_name(reserved_index, custom_index, name):
        # search through reserved names first
        if name in reserved_index:
            return reserved_index[name]
        elif name in custom_index:
            return custom_index[name]
        else:
            raise KeyError(f"There is no metadata with name '{name}'")

    def get_by_name(
        self, name: str
    ) -> Union[DataRowMetadataSchema, Dict[str, DataRowMetadataSchema]]:
        """ Get metadata by name

        >>> mdo.get_by_name(name)

        Args:
            name (str): Name of metadata schema

        Returns:
            Metadata schema as `DataRowMetadataSchema` or dict, in case of Enum metadata

        Raises:
            KeyError: When provided name is not presented in neither reserved nor custom metadata list
        """
        return self._lookup_in_index_by_name(self.reserved_by_name,
                                             self.custom_by_name, name)

    def _get_by_name_normalized(self, name: str) -> DataRowMetadataSchema:
        """ Get metadata by name. For options, it provides the option schema instead of list of 
        options 
        """
        # using `normalized` indices to find options by name as well
        return self._lookup_in_index_by_name(self.reserved_by_name_normalized,
                                             self.custom_by_name_normalized,
                                             name)

    @staticmethod
    def _make_name_index(
        fields: List[DataRowMetadataSchema]
    ) -> Dict[str, Union[DataRowMetadataSchema, Dict[str,
                                                     DataRowMetadataSchema]]]:
        index = {}
        for f in fields:
            if f.options:
                index[f.name] = {}
                for o in f.options:
                    index[f.name][o.name] = o
            else:
                index[f.name] = f
        return index

    @staticmethod
    def _make_normalized_name_index(
        fields: List[DataRowMetadataSchema]
    ) -> Dict[str, DataRowMetadataSchema]:
        index = {}
        for f in fields:
            index[f.name] = f
        return index

    @staticmethod
    def _make_id_index(
        fields: List[DataRowMetadataSchema]
    ) -> Dict[SchemaId, DataRowMetadataSchema]:
        index = {}
        for f in fields:
            index[f.uid] = f
            if f.options:
                for o in f.options:
                    index[o.uid] = o
        return index

    def _get_ontology(self) -> List[Dict[str, Any]]:
        query = """query GetMetadataOntologyBetaPyApi {
        customMetadataOntology {
                id
                name
                kind
                reserved
                options {
                    id
                    kind
                    name
                    reserved
                }
        }}
        """
        return self._client.execute(query)["customMetadataOntology"]

    @staticmethod
    def _parse_ontology(raw_ontology) -> List[DataRowMetadataSchema]:
        fields = []
        copy = deepcopy(raw_ontology)
        for schema in copy:
            schema["uid"] = schema["id"]
            options = None
            if schema.get("options"):
                options = []
                for option in schema["options"]:
                    option["uid"] = option["id"]
                    options.append(
                        DataRowMetadataSchema(**{
                            **option,
                            **{
                                "parent": schema["uid"]
                            }
                        }))
            schema["options"] = options
            fields.append(DataRowMetadataSchema(**schema))

        return fields

    def refresh_ontology(self):
        """ Update the `DataRowMetadataOntology` instance with the latest
            metadata ontology schemas
        """
        self._raw_ontology = self._get_ontology()
        self._build_ontology()

    def create_schema(self,
                      name: str,
                      kind: DataRowMetadataKind,
                      options: List[str] = None) -> DataRowMetadataSchema:
        """ Create metadata schema

        >>> mdo.create_schema(name, kind, options)

        Args:
            name (str): Name of metadata schema
            kind (DataRowMetadataKind): Kind of metadata schema as `DataRowMetadataKind`
            options (List[str]): List of Enum options

        Returns:
            Created metadata schema as `DataRowMetadataSchema`

        Raises:
            KeyError: When provided name is not a valid custom metadata
        """
        if not isinstance(kind, DataRowMetadataKind):
            raise ValueError(f"kind '{kind}' must be a `DataRowMetadataKind`")

        upsert_schema = _UpsertCustomMetadataSchemaInput(name=name,
                                                         kind=kind.value)
        if options:
            if kind != DataRowMetadataKind.enum:
                raise ValueError(
                    f"Kind '{kind}' must be an Enum, if Enum options are provided"
                )
            upsert_enum_options = [
                _UpsertCustomMetadataSchemaEnumOptionInput(
                    name=o, kind=DataRowMetadataKind.option.value)
                for o in options
            ]
            upsert_schema.options = upsert_enum_options

        return self._upsert_schema(upsert_schema)

    def update_schema(self, name: str, new_name: str) -> DataRowMetadataSchema:
        """ Update metadata schema

        >>> mdo.update_schema(name, new_name)

        Args:
            name (str): Current name of metadata schema
            new_name (str): New name of metadata schema

        Returns:
            Updated metadata schema as `DataRowMetadataSchema`

        Raises:
            KeyError: When provided name is not a valid custom metadata
        """
        schema = self._validate_custom_schema_by_name(name)
        upsert_schema = _UpsertCustomMetadataSchemaInput(id=schema.uid,
                                                         name=new_name,
                                                         kind=schema.kind.value)
        if schema.options:
            upsert_enum_options = [
                _UpsertCustomMetadataSchemaEnumOptionInput(
                    id=o.uid,
                    name=o.name,
                    kind=DataRowMetadataKind.option.value)
                for o in schema.options
            ]
            upsert_schema.options = upsert_enum_options

        return self._upsert_schema(upsert_schema)

    def update_enum_option(self, name: str, option: str,
                           new_option: str) -> DataRowMetadataSchema:
        """ Update Enum metadata schema option

        >>> mdo.update_enum_option(name, option, new_option)

        Args:
            name (str): Name of metadata schema to update
            option (str): Name of Enum option to update
            new_option (str): New name of Enum option

        Returns:
            Updated metadata schema as `DataRowMetadataSchema`

        Raises:
            KeyError: When provided name is not a valid custom metadata
        """
        schema = self._validate_custom_schema_by_name(name)
        if schema.kind != DataRowMetadataKind.enum:
            raise ValueError(
                f"Updating Enum option is only supported for Enum metadata schema"
            )
        valid_options: List[str] = [o.name for o in schema.options]

        if option not in valid_options:
            raise ValueError(
                f"Enum option '{option}' is not a valid option for Enum '{name}', valid options are: {valid_options}"
            )
        upsert_schema = _UpsertCustomMetadataSchemaInput(id=schema.uid,
                                                         name=schema.name,
                                                         kind=schema.kind.value)
        upsert_enum_options = []
        for o in schema.options:
            enum_option = _UpsertCustomMetadataSchemaEnumOptionInput(
                id=o.uid, name=o.name, kind=o.kind.value)
            if enum_option.name == option:
                enum_option.name = new_option
            upsert_enum_options.append(enum_option)
        upsert_schema.options = upsert_enum_options

        return self._upsert_schema(upsert_schema)

    def delete_schema(self, name: str) -> bool:
        """ Delete metadata schema

        >>> mdo.delete_schema(name)

        Args:
            name: Name of metadata schema to delete

        Returns:
            True if deletion is successful, False if unsuccessful

        Raises:
            KeyError: When provided name is not a valid custom metadata
        """
        schema = self._validate_custom_schema_by_name(name)
        query = """mutation DeleteCustomMetadataSchemaPyApi($where: WhereUniqueIdInput!) {
                deleteCustomMetadataSchema(schema: $where){
                    success
                }
            }"""
        res = self._client.execute(query, {'where': {
            'id': schema.uid
        }})['deleteCustomMetadataSchema']
        self.refresh_ontology()

        return res['success']

    def parse_metadata(
        self, unparsed: List[Dict[str,
                                  List[Union[str,
                                             Dict]]]]) -> List[DataRowMetadata]:
        """ Parse metadata responses

        >>> mdo.parse_metadata([metadata])

        Args:
            unparsed: An unparsed metadata export

        Returns:
            metadata: List of `DataRowMetadata`
        """
        parsed = []
        if isinstance(unparsed, dict):
            raise ValueError("Pass a list of dictionaries")

        for dr in unparsed:
            fields = []
            if "fields" in dr:
                fields = self.parse_metadata_fields(dr["fields"])
            parsed.append(
                DataRowMetadata(data_row_id=dr["dataRowId"],
                                global_key=dr["globalKey"],
                                fields=fields))
        return parsed

    def parse_metadata_fields(
            self, unparsed: List[Dict[str,
                                      Dict]]) -> List[DataRowMetadataField]:
        """ Parse metadata fields as list of `DataRowMetadataField`

        >>> mdo.parse_metadata_fields([metadata_fields])

        Args:
            unparsed: An unparsed list of metadata represented as a dict containing 'schemaId' and 'value'

        Returns:
            metadata: List of `DataRowMetadataField`
        """
        parsed = []
        if isinstance(unparsed, dict):
            raise ValueError("Pass a list of dictionaries")

        for f in unparsed:
            if f["schemaId"] not in self.fields_by_id:
                # Fetch latest metadata ontology if metadata can't be found
                self.refresh_ontology()
                if f["schemaId"] not in self.fields_by_id:
                    raise ValueError(
                        f"Schema Id `{f['schemaId']}` not found in ontology")

            schema = self.fields_by_id[f["schemaId"]]
            if schema.kind == DataRowMetadataKind.enum:
                continue
            elif schema.kind == DataRowMetadataKind.option:
                field = DataRowMetadataField(schema_id=schema.parent,
                                             value=schema.uid)
            elif schema.kind == DataRowMetadataKind.datetime:
                field = DataRowMetadataField(schema_id=schema.uid,
                                             value=format_iso_from_string(
                                                 f["value"]))
            else:
                field = DataRowMetadataField(schema_id=schema.uid,
                                             value=f["value"])

            field.name = schema.name
            parsed.append(field)
        return parsed

    def bulk_upsert(
            self, metadata: List[DataRowMetadata]
    ) -> List[DataRowMetadataBatchResponse]:
        """Upsert metadata to a list of data rows
        
        You may specify data row by either data_row_id or global_key

        >>> metadata = DataRowMetadata(
        >>>                 data_row_id="datarow-id", # Alternatively, set global_key="global-key"
        >>>                 fields=[
        >>>                        DataRowMetadataField(schema_id="schema-id", value="my-message"),
        >>>                        ...
        >>>                    ]
        >>>    )
        >>> mdo.batch_upsert([metadata])

        Args:
            metadata: List of DataRow Metadata to upsert

        Returns:
            list of unsuccessful upserts.
            An empty list means the upload was successful.
        """

        if not len(metadata):
            raise ValueError("Empty list passed")

        def _batch_upsert(
            upserts: List[_UpsertBatchDataRowMetadata]
        ) -> List[DataRowMetadataBatchResponse]:
            query = """mutation UpsertDataRowMetadataBetaPyApi($metadata: [DataRowCustomMetadataBatchUpsertInput!]!) {
                upsertDataRowCustomMetadata(data: $metadata){
                    globalKey
                    dataRowId
                    error
                    fields {
                        value
                        schemaId
                    }
                }
            }"""
            res = self._client.execute(
                query, {"metadata": upserts})['upsertDataRowCustomMetadata']
            return [
                DataRowMetadataBatchResponse(global_key=r['globalKey'],
                                             data_row_id=r['dataRowId'],
                                             error=r['error'],
                                             fields=self.parse_metadata(
                                                 [r])[0].fields) for r in res
            ]

        items = []
        for m in metadata:
            items.append(
                _UpsertBatchDataRowMetadata(
                    global_key=m.global_key,
                    data_row_id=m.data_row_id,
                    fields=list(
                        chain.from_iterable(
                            self._parse_upsert(f, m.data_row_id)
                            for f in m.fields))).dict(by_alias=True))
        res = _batch_operations(_batch_upsert, items, self._batch_size)
        return res

    def bulk_delete(
        self, deletes: List[DeleteDataRowMetadata]
    ) -> List[DataRowMetadataBatchResponse]:
        """ Delete metadata from a datarow by specifiying the fields you want to remove

        >>> delete = DeleteDataRowMetadata(
        >>>                 data_row_id=UniqueId("datarow-id"),
        >>>                 fields=[
        >>>                        "schema-id-1",
        >>>                        "schema-id-2"
        >>>                        ...
        >>>                    ]
        >>>    )
        >>> mdo.batch_delete([metadata])

        >>> delete = DeleteDataRowMetadata(
        >>>                 data_row_id=GlobalKey("global-key"),
        >>>                 fields=[
        >>>                        "schema-id-1",
        >>>                        "schema-id-2"
        >>>                        ...
        >>>                    ]
        >>>    )
        >>> mdo.batch_delete([metadata])

        >>> delete = DeleteDataRowMetadata(
        >>>                 data_row_id="global-key",
        >>>                 fields=[
        >>>                        "schema-id-1",
        >>>                        "schema-id-2"
        >>>                        ...
        >>>                    ]
        >>>    )
        >>> mdo.batch_delete([metadata])


        Args:
            deletes: Data row and schema ids to delete
                For data row, we support UniqueId, str, and GlobalKey. 
                If you pass a str, we will assume it is a UniqueId
                Do not pass a mix of data row ids and global keys in the same list

        Returns:
            list of unsuccessful deletions.
            An empty list means all data rows were successfully deleted.

        """

        if not len(deletes):
            raise ValueError("The 'deletes' list cannot be empty.")

        passed_strings = False
        for i, delete in enumerate(deletes):
            if isinstance(delete.data_row_id, str):
                passed_strings = True
                deletes[i] = DeleteDataRowMetadata(data_row_id=UniqueId(
                    delete.data_row_id),
                                                   fields=delete.fields)
            elif isinstance(delete.data_row_id, UniqueId):
                continue
            elif isinstance(delete.data_row_id, GlobalKey):
                continue
            else:
                raise ValueError(
                    f"Invalid data row identifier type '{type(delete.data_row_id)}' for '{delete.data_row_id}'"
                )

            if passed_strings:
                warnings.warn(
                    "Using string for data row id will be deprecated. Please use "
                    "UniqueId instead.")

        def _batch_delete(
            deletes: List[_DeleteBatchDataRowMetadata]
        ) -> List[DataRowMetadataBatchResponse]:
            query = """mutation DeleteDataRowMetadataBetaPyApi($deletes: [DataRowIdentifierCustomMetadataBatchDeleteInput!]) {
                deleteDataRowCustomMetadata(dataRowIdentifiers: $deletes) {
                    dataRowId
                    error
                    fields {
                        value
                        schemaId
                    }
                }
            }
            """
            res = self._client.execute(
                query, {"deletes": deletes})['deleteDataRowCustomMetadata']
            failures = []
            for dr in res:
                dr['fields'] = [f['schemaId'] for f in dr['fields']]
                failures.append(DataRowMetadataBatchResponse(**dr))
            return failures

        items = [self._validate_delete(m) for m in deletes]
        return _batch_operations(_batch_delete,
                                 items,
                                 batch_size=self._batch_size)

    @overload
    def bulk_export(self, data_row_ids: List[str]) -> List[DataRowMetadata]:
        pass

    @overload
    def bulk_export(self,
                    data_row_ids: DataRowIdentifiers) -> List[DataRowMetadata]:
        pass

    def bulk_export(self, data_row_ids) -> List[DataRowMetadata]:
        """ Exports metadata for a list of data rows

        >>> mdo.bulk_export([data_row.uid for data_row in data_rows])

        Args:
            data_row_ids: List of data data rows to fetch metadata for. This can be a list of strings or a DataRowIdentifiers object
            DataRowIdentifier objects are lists of ids or global keys. A DataIdentifier object can be a UniqueIds or GlobalKeys class.
        Returns:
            A list of DataRowMetadata.
            There will be one DataRowMetadata for each data_row_id passed in.
            This is true even if the data row does not have any meta data.
            Data rows without metadata will have empty `fields`.

        """
        if not len(data_row_ids):
            raise ValueError("Empty list passed")

        if isinstance(data_row_ids,
                      list) and len(data_row_ids) > 0 and isinstance(
                          data_row_ids[0], str):
            data_row_ids = UniqueIds(data_row_ids)
            warnings.warn("Using data row ids will be deprecated. Please use "
                          "UniqueIds or GlobalKeys instead.")

        def _bulk_export(
                _data_row_ids: DataRowIdentifiers) -> List[DataRowMetadata]:
            query = """query dataRowCustomMetadataPyApi($dataRowIdentifiers: DataRowCustomMetadataDataRowIdentifiersInput) {
                dataRowCustomMetadata(where: {dataRowIdentifiers : $dataRowIdentifiers}) {
                    dataRowId
                    globalKey
                    fields {
                        value
                        schemaId
                    }
                }
            }
            """
            return self.parse_metadata(
                self._client.execute(
                    query, {
                        "dataRowIdentifiers": {
                            "ids": [id for id in _data_row_ids],
                            "idType": _data_row_ids.id_type
                        }
                    })['dataRowCustomMetadata'])

        return _batch_operations(_bulk_export,
                                 data_row_ids,
                                 batch_size=self._batch_size)

    def parse_upsert_metadata(self, metadata_fields) -> List[Dict[str, Any]]:
        """ Converts either `DataRowMetadataField` or a dictionary representation
            of `DataRowMetadataField` into a validated, flattened dictionary of
            metadata fields that are used to create data row metadata. Used
            internally in `Dataset.create_data_rows()`

        Args:
            metadata_fields: List of `DataRowMetadataField` or a dictionary representation
                of `DataRowMetadataField`
        Returns:
            List of dictionaries representing a flattened view of metadata fields
        """

        def _convert_metadata_field(metadata_field):
            if isinstance(metadata_field, DataRowMetadataField):
                return metadata_field
            elif isinstance(metadata_field, dict):
                if not "value" in metadata_field:
                    raise ValueError(
                        f"Custom metadata field '{metadata_field}' must have a 'value' key"
                    )
                if not "schema_id" in metadata_field and not "name" in metadata_field:
                    raise ValueError(
                        f"Custom metadata field '{metadata_field}' must have either 'schema_id' or 'name' key"
                    )
                return DataRowMetadataField(
                    schema_id=metadata_field.get("schema_id"),
                    name=metadata_field.get("name"),
                    value=metadata_field["value"])
            else:
                raise ValueError(
                    f"Metadata field '{metadata_field}' is neither 'DataRowMetadataField' type or a dictionary"
                )

        # Convert all metadata fields to DataRowMetadataField type
        metadata_fields = [_convert_metadata_field(m) for m in metadata_fields]
        parsed_metadata = list(
            chain.from_iterable(self._parse_upsert(m) for m in metadata_fields))
        return [m.dict(by_alias=True) for m in parsed_metadata]

    def _upsert_schema(
        self, upsert_schema: _UpsertCustomMetadataSchemaInput
    ) -> DataRowMetadataSchema:
        query = """mutation UpsertCustomMetadataSchemaPyApi($data: UpsertCustomMetadataSchemaInput!) {
                upsertCustomMetadataSchema(data: $data){
                    id
                    name
                    kind
                    options {
                        id
                        name
                        kind
                    }
                }
            }"""
        res = self._client.execute(
            query, {"data": upsert_schema.dict(exclude_none=True)
                   })['upsertCustomMetadataSchema']
        self.refresh_ontology()
        return _parse_metadata_schema(res)

    def _load_option_by_name(self, metadatum: DataRowMetadataField):
        is_value_a_valid_schema_id = metadatum.value in self.fields_by_id
        if not is_value_a_valid_schema_id:
            metadatum_by_name = self.get_by_name(metadatum.name)
            if metadatum.value not in metadatum_by_name:
                raise KeyError(
                    f"There is no enum option by name '{metadatum.value}' for enum name '{metadatum.name}'"
                )
            metadatum.value = metadatum_by_name[metadatum.value].uid

    def _load_schema_id_by_name(self, metadatum: DataRowMetadataField):
        """
        Loads schema id by name for a metadata field including options schema id.
        """
        if metadatum.name is None:
            return

        if metadatum.schema_id is None:
            schema = self._get_by_name_normalized(metadatum.name)
            metadatum.schema_id = schema.uid
            if schema.options:
                self._load_option_by_name(metadatum)

    def _parse_upsert(
            self,
            metadatum: DataRowMetadataField,
            data_row_id: Optional[str] = None
    ) -> List[_UpsertDataRowMetadataInput]:
        """Format for metadata upserts to GQL"""

        self._load_schema_id_by_name(metadatum)

        if metadatum.schema_id not in self.fields_by_id:
            # Fetch latest metadata ontology if metadata can't be found
            self.refresh_ontology()
            if metadatum.schema_id not in self.fields_by_id:
                raise ValueError(
                    f"Schema Id `{metadatum.schema_id}` not found in ontology")

        schema = self.fields_by_id[metadatum.schema_id]
        try:
            if schema.kind == DataRowMetadataKind.datetime:
                parsed = _validate_parse_datetime(metadatum)
            elif schema.kind == DataRowMetadataKind.string:
                parsed = _validate_parse_text(metadatum)
            elif schema.kind == DataRowMetadataKind.number:
                parsed = _validate_parse_number(metadatum)
            elif schema.kind == DataRowMetadataKind.embedding:
                parsed = _validate_parse_embedding(metadatum)
            elif schema.kind == DataRowMetadataKind.enum:
                parsed = _validate_enum_parse(schema, metadatum)
            elif schema.kind == DataRowMetadataKind.option:
                raise ValueError(
                    "An Option id should not be set as the Schema id")
            else:
                raise ValueError(f"Unknown type: {schema}")
        except ValueError as e:
            error_str = f"Could not validate metadata [{metadatum}]"
            if data_row_id:
                error_str += f", data_row_id='{data_row_id}'"
            raise ValueError(f"{error_str}. Reason: {e}")

        return [_UpsertDataRowMetadataInput(**p) for p in parsed]

    def _validate_delete(self, delete: DeleteDataRowMetadata):
        if not len(delete.fields):
            raise ValueError(f"No fields specified for {delete.data_row_id}")

        deletes = set()
        for schema_id in delete.fields:
            if schema_id not in self.fields_by_id:
                # Fetch latest metadata ontology if metadata can't be found
                self.refresh_ontology()
                if schema_id not in self.fields_by_id:
                    raise ValueError(
                        f"Schema Id `{schema_id}` not found in ontology")

            schema = self.fields_by_id[schema_id]
            # handle users specifying enums by adding all option enums
            if schema.kind == DataRowMetadataKind.enum:
                [deletes.add(o.uid) for o in schema.options]

            deletes.add(schema.uid)

        return _DeleteBatchDataRowMetadata(
            data_row_identifier=delete.data_row_id,
            schema_ids=list(delete.fields)).dict(by_alias=True)

    def _validate_custom_schema_by_name(self,
                                        name: str) -> DataRowMetadataSchema:
        if name not in self.custom_by_name_normalized:
            # Fetch latest metadata ontology if metadata can't be found
            self.refresh_ontology()
            if name not in self.custom_by_name_normalized:
                raise KeyError(f"'{name}' is not a valid custom metadata")

        return self.custom_by_name_normalized[name]


def _batch_items(iterable: List[Any], size: int) -> Generator[Any, None, None]:
    l = len(iterable)
    for ndx in range(0, l, size):
        yield iterable[ndx:min(ndx + size, l)]


def _batch_operations(
    batch_function: _BatchFunction,
    items: List,
    batch_size: int = 100,
):
    response = []

    for batch in _batch_items(items, batch_size):
        response += batch_function(batch)
    return response


def _validate_parse_embedding(
        field: DataRowMetadataField
) -> List[Dict[str, Union[SchemaId, Embedding]]]:

    if isinstance(field.value, list):
        if not (Embedding.min_items <= len(field.value) <= Embedding.max_items):
            raise ValueError(
                "Embedding length invalid. "
                "Must have length within the interval "
                f"[{Embedding.min_items},{Embedding.max_items}]. Found {len(field.value)}"
            )
        field.value = [float(x) for x in field.value]
    else:
        raise ValueError(
            f"Expected a list for embedding. Found {type(field.value)}")
    return [field.dict(by_alias=True)]


def _validate_parse_number(
    field: DataRowMetadataField
) -> List[Dict[str, Union[SchemaId, str, float, int]]]:
    field.value = float(field.value)
    return [field.dict(by_alias=True)]


def _validate_parse_datetime(
        field: DataRowMetadataField) -> List[Dict[str, Union[SchemaId, str]]]:
    if isinstance(field.value, str):
        field.value = format_iso_from_string(field.value)
    elif not isinstance(field.value, datetime):
        raise TypeError(
            f"Value for datetime fields must be either a string or datetime object. Found {type(field.value)}"
        )

    return [{
        "schemaId": field.schema_id,
        "value": format_iso_datetime(field.value)
    }]


def _validate_parse_text(
        field: DataRowMetadataField) -> List[Dict[str, Union[SchemaId, str]]]:
    if not isinstance(field.value, str):
        raise ValueError(
            f"Expected a string type for the text field. Found {type(field.value)}"
        )

    if len(field.value) > String.max_length:
        raise ValueError(
            f"String fields cannot exceed {String.max_length} characters.")

    return [field.dict(by_alias=True)]


def _validate_enum_parse(
        schema: DataRowMetadataSchema,
        field: DataRowMetadataField) -> List[Dict[str, Union[SchemaId, dict]]]:
    if schema.options:
        if field.value not in {o.uid for o in schema.options}:
            raise ValueError(
                f"Option `{field.value}` not found for {field.schema_id}")
    else:
        raise ValueError("Incorrectly specified enum schema")

    return [{
        "schemaId": field.schema_id,
        "value": {}
    }, {
        "schemaId": field.value,
        "value": {}
    }]


def _parse_metadata_schema(
        unparsed: Dict[str, Union[str, List]]) -> DataRowMetadataSchema:
    uid = unparsed['id']
    name = unparsed['name']
    kind = DataRowMetadataKind(unparsed['kind'])
    options = [
        DataRowMetadataSchema(uid=o['id'],
                              name=o['name'],
                              reserved=False,
                              kind=DataRowMetadataKind.option,
                              parent=uid) for o in unparsed['options']
    ]
    return DataRowMetadataSchema(uid=uid,
                                 name=name,
                                 reserved=False,
                                 kind=kind,
                                 options=options or None)
